@@ -14,6 +14,7 @@ from devpi_common.types import cached_property
 from distutils.util import strtobool
 from functools import partial
 from .log import threadlog
+from . import fileutil
 from . import hookspecs
 import json
 import devpi_server
@@ -215,6 +216,11 @@ def add_replica_options(parser, pluginmanager):
         help="Number of seconds to wait before proxied requests from "
              "the replica to the master time out (login, uploads etc).")
 
+    parser.addoption(
+        "--no-replica-streaming", dest="replica_streaming",
+        default=True, action="store_false",
+        help="use separate requests instead of replica streaming protocol")
+
 
 def add_request_options(parser, pluginmanager):
     parser.addoption(
@@ -413,12 +419,12 @@ def get_parser(pluginmanager):
 
 
 def find_config_file():
-    import appdirs
-    config_dirs = appdirs.site_config_dir(
+    import platformdirs
+    config_dirs = platformdirs.site_config_dir(
         'devpi-server', 'devpi', multipath=True)
     config_dirs = config_dirs.split(os.pathsep)
     config_dirs.append(
-        appdirs.user_config_dir('devpi-server', 'devpi'))
+        platformdirs.user_config_dir('devpi-server', 'devpi'))
     config_files = []
     for config_dir in config_dirs:
         config_file = os.path.join(config_dir, 'devpi-server.yml')
@@ -673,11 +679,7 @@ class Config(object):
         prefix = "-" + self.path_nodeinfo.basename
         with NamedTemporaryFile(prefix=prefix, delete=False, dir=nodeinfo_dir.strpath) as f:
             f.write(json.dumps(self.nodeinfo, indent=2).encode('utf-8'))
-        try:
-            os.remove(self.path_nodeinfo.strpath)
-        except FileNotFoundError:
-            pass
-        os.rename(f.name, self.path_nodeinfo.strpath)
+        fileutil.rename(f.name, self.path_nodeinfo.strpath)
         threadlog.info("wrote nodeinfo to: %s", self.path_nodeinfo)
 
     @property
@@ -744,6 +746,10 @@ class Config(object):
     def replica_max_retries(self):
         return getattr(self.args, 'replica_max_retries', None)
 
+    @property
+    def replica_streaming(self):
+        return getattr(self.args, 'replica_streaming', True)
+
     @property
     def requests_only(self):
         return getattr(self.args, 'requests_only', False)
