@@ -12,11 +12,12 @@ import contextlib
 import py
 from . import mythread
 from .fileutil import loads
-from .interfaces import IStorageConnection2
+from .interfaces import IStorageConnection3
 from .log import threadlog, thread_push_log, thread_pop_log
 from .log import thread_change_log_prefix
-from .readonly import get_mutable_deepcopy, ensure_deeply_readonly, \
-                      is_deeply_readonly
+from .readonly import ensure_deeply_readonly
+from .readonly import get_mutable_deepcopy
+from .readonly import is_deeply_readonly
 from .filestore import FileEntry
 from .fileutil import read_int_from_file, write_int_to_file
 import attr
@@ -223,7 +224,7 @@ class KeyFS(object):
         self._keys = {}
         self._threadlocal = mythread.threading.local()
         self._cv_new_transaction = mythread.threading.Condition()
-        self._import_subscriber = {}
+        self._import_subscriber = None
         self.notifier = TxNotificationThread(self)
         self._storage = storage(
             self.basedir,
@@ -232,7 +233,7 @@ class KeyFS(object):
         self._readonly = readonly
 
     def get_connection(self, closing=True, write=False):
-        conn = IStorageConnection2(
+        conn = IStorageConnection3(
             self._storage.get_connection(closing=False, write=write))
         if closing:
             return contextlib.closing(conn)
@@ -242,38 +243,24 @@ class KeyFS(object):
         self._storage.perform_crash_recovery()
 
     def import_changes(self, serial, changes):
-        subscriber_task_infos = []
         with self.get_connection(write=True) as conn:
             with conn.write_transaction() as fswriter:
                 next_serial = conn.last_changelog_serial + 1
                 assert next_serial == serial, (next_serial, serial)
-                for relpath, tup in changes.items():
-                    keyname, back_serial, val = tup
-                    typedkey = self.get_key_instance(keyname, relpath)
-                    try:
-                        fswriter.record_set(
-                            typedkey, get_mutable_deepcopy(val),
-                            back_serial=back_serial)
-                    except TypeError as e:
-                        # for backward compatibility with storage backends
-                        # which didn't support the back_serial keyword
-                        if not len(e.args) or 'back_serial' not in e.args[0]:
-                            raise
-                        fswriter.record_set(
-                            typedkey, get_mutable_deepcopy(val))
-                    meth = self._import_subscriber.get(keyname)
-                    if meth is not None:
-                        subscriber_task_infos.append(
-                            (meth, typedkey, val, back_serial))
-                if subscriber_task_infos:
-                    with self.transaction(write=False, at_serial=serial):
-                        for meth, typedkey, val, back_serial in subscriber_task_infos:
-                            threadlog.debug("calling import subscriber %r", meth)
-                            meth(fswriter.conn, serial, typedkey, val, back_serial)
+                changes = {
+                    self.get_key_instance(keyname, relpath): (val, back_serial)
+                    for relpath, (keyname, back_serial, val) in changes.items()}
+                for typedkey, (val, back_serial) in changes.items():
+                    fswriter.record_set(
+                        typedkey, get_mutable_deepcopy(val),
+                        back_serial=back_serial)
+        if callable(self._import_subscriber):
+            with self.transaction(write=False, at_serial=serial):
+                self._import_subscriber(serial, changes)
 
-    def subscribe_on_import(self, key, subscriber):
-        assert key.name not in self._import_subscriber
-        self._import_subscriber[key.name] = subscriber
+    def subscribe_on_import(self, subscriber):
+        assert self._import_subscriber is None
+        self._import_subscriber = subscriber
 
     def _notify_on_commit(self, serial):
         self.release_all_wait_tx()
@@ -402,7 +389,7 @@ class KeyFS(object):
         tx = self.begin_transaction_in_thread(write=write, at_serial=at_serial)
         try:
             yield tx
-        except:
+        except BaseException:
             self.rollback_transaction_in_thread()
             raise
         self.commit_transaction_in_thread()
@@ -441,7 +428,7 @@ class PTypedKey:
         self.keyfs.notifier.on_key_change(self.name, callback)
 
     def __repr__(self):
-        return "<PTypedKey %r type %r>" %(self.pattern, self.type.__name__)
+        return f"<PTypedKey {self.pattern!r} type {self.type.__name__!r}>"
 
 
 class KeyChangeEvent:
@@ -469,7 +456,7 @@ class TypedKey:
         return self.relpath == other.relpath
 
     def __repr__(self):
-        return "<TypedKey %r type %r>" %(self.relpath, self.type.__name__)
+        return f"<TypedKey {self.name} {self.type.__name__} {self.relpath}>"
 
     def get(self, readonly=True):
         return self.keyfs.tx.get(self, readonly=readonly)
@@ -493,9 +480,9 @@ class TypedKey:
 
     def set(self, val):
         if not isinstance(val, self.type):
-            raise TypeError("%r requires value of type %r, got %r" %(
-                            self.relpath, self.type.__name__,
-                            type(val).__name__))
+            raise TypeError(
+                "%r requires value of type %r, got %r" % (
+                    self.relpath, self.type.__name__, type(val).__name__))
         self.keyfs.tx.set(self, val)
 
     def exists(self):
@@ -531,6 +518,25 @@ def get_relpath_at(self, relpath, serial):
     raise KeyError(relpath)
 
 
+def io_file_new_open(self, path):
+    """ Fallback method for legacy storage connections. """
+    from tempfile import TemporaryFile
+    return TemporaryFile()
+
+
+def io_file_set(self, path, content_or_file, _io_file_set):
+    """ Fallback method wrapper for legacy storage connections. """
+    # _io_file_set is from the original class
+    if not isinstance(content_or_file, bytes):
+        content_or_file.seek(0)
+        content_or_file = content_or_file.read()
+    if len(content_or_file) > 1048576:
+        threadlog.warn(
+            "Got content with %.1f megabytes in memory while setting content for %s",
+            len(content_or_file) / 1048576, path)
+    return _io_file_set(self, path, content_or_file)
+
+
 def iter_serial_and_value_backwards(conn, relpath, last_serial):
     while last_serial >= 0:
         tup = conn.get_changes(last_serial).get(relpath)
@@ -723,7 +729,10 @@ class Transaction(object):
                     val = self.cache.get(typedkey)
                     # None signals deletion
                     fswriter.record_set(typedkey, val)
-                commit_serial = self.conn.last_changelog_serial + 1
+                commit_serial = getattr(fswriter, "commit_serial", notset)
+                if commit_serial is notset:
+                    # for storages which don't have the attribute yet
+                    commit_serial = self.conn.last_changelog_serial + 1
         finally:
             self._close()
         self.commit_serial = commit_serial
