@@ -5,8 +5,10 @@ from .interfaces import IStorageConnection2
 from .keyfs import RelpathInfo
 from .keyfs import get_relpath_at
 from .log import threadlog, thread_push_log, thread_pop_log
+from .mythread import current_thread
 from .readonly import ReadonlyView
 from .readonly import ensure_deeply_readonly, get_mutable_deepcopy
+from .sizeof import gettotalsizeof
 from repoze.lru import LRUCache
 from zope.interface import implementer
 import contextlib
@@ -20,41 +22,64 @@ absent = object()
 
 
 class BaseConnection:
+    _get_relpath_at = get_relpath_at
+
     def __init__(self, sqlconn, basedir, storage):
         self._sqlconn = sqlconn
         self._basedir = basedir
         self.dirty_files = {}
         self.storage = storage
         self._changelog_cache = storage._changelog_cache
+        self._relpath_cache = storage._relpath_cache
 
     def _explain(self, query, *args):
         # for debugging
         c = self._sqlconn.cursor()
         r = c.execute("EXPLAIN " + query, *args)
-        return r.fetchall()
+        result = r.fetchall()
+        c.close()
+        return result
 
     def _explain_query_plan(self, query, *args):
         # for debugging
         c = self._sqlconn.cursor()
         r = c.execute("EXPLAIN QUERY PLAN " + query, *args)
-        return r.fetchall()
+        result = r.fetchall()
+        c.close()
+        return result
+
+    def _print_rows(self, rows):
+        # for debugging
+        for row in rows:
+            print(row)
 
     def fetchall(self, query, *args):
         c = self._sqlconn.cursor()
         # print(query)
-        # pprint(self._explain(query, *args))
-        # for row in self._explain_query_plan(query, *args):
-        #     print(row)
+        # self._print_rows(self._explain(query, *args))
+        # self._print_rows(self._explain_query_plan(query, *args))
         r = c.execute(query, *args)
-        return r.fetchall()
+        result = r.fetchall()
+        c.close()
+        return result
+
+    def fetchone(self, query, *args):
+        c = self._sqlconn.cursor()
+        # print(query)
+        # self._print_rows(self._explain(query, *args))
+        # self._print_rows(self._explain_query_plan(query, *args))
+        r = c.execute(query, *args)
+        result = r.fetchone()
+        c.close()
+        return result
 
     def iterall(self, query, *args):
         c = self._sqlconn.cursor()
         # print(query)
-        # pprint(self._explain(query, *args))
-        # for row in self._explain_query_plan(query, *args):
-        #     print(row)
-        return c.execute(query, *args)
+        # self._print_rows(self._explain(query, *args))
+        # self._print_rows(self._explain_query_plan(query, *args))
+        yield from c.execute(query, *args)
+        c.close()
 
     def close(self):
         self._sqlconn.close()
@@ -71,38 +96,37 @@ class BaseConnection:
 
     def db_read_last_changelog_serial(self):
         q = 'SELECT MAX(_ROWID_) FROM "changelog" LIMIT 1'
-        res = self._sqlconn.execute(q).fetchone()[0]
+        res = self.fetchone(q)[0]
         return -1 if res is None else res
 
     def db_read_typedkey(self, relpath):
         q = "SELECT keyname, serial FROM kv WHERE key = ?"
-        c = self._sqlconn.cursor()
-        row = c.execute(q, (relpath,)).fetchone()
+        row = self.fetchone(q, (relpath,))
         if row is None:
             raise KeyError(relpath)
         return tuple(row[:2])
 
     def db_write_typedkey(self, relpath, name, next_serial):
         q = "INSERT OR REPLACE INTO kv (key, keyname, serial) VALUES (?, ?, ?)"
-        self._sqlconn.execute(q, (relpath, name, next_serial))
+        self.fetchone(q, (relpath, name, next_serial))
 
     def write_changelog_entry(self, serial, entry):
         threadlog.debug("writing changelog for serial %s", serial)
         data = dumps(entry)
-        self._sqlconn.execute(
+        self.fetchone(
             "INSERT INTO changelog (serial, data) VALUES (?, ?)",
             (serial, sqlite3.Binary(data)))
 
     def get_raw_changelog_entry(self, serial):
         q = "SELECT data FROM changelog WHERE serial = ?"
-        row = self._sqlconn.execute(q, (serial,)).fetchone()
+        row = self.fetchone(q, (serial,))
         if row is not None:
             return bytes(row[0])
         return None
 
     def get_changes(self, serial):
-        changes = self._changelog_cache.get(serial)
-        if changes is None:
+        changes = self._changelog_cache.get(serial, absent)
+        if changes is absent:
             data = self.get_raw_changelog_entry(serial)
             changes, rel_renames = loads(data)
             # make values in changes read only so no calling site accidentally
@@ -113,15 +137,23 @@ class BaseConnection:
         return changes
 
     def get_relpath_at(self, relpath, serial):
-        result = self._changelog_cache.get((serial, relpath), absent)
+        result = self._relpath_cache.get((serial, relpath), absent)
+        if result is absent:
+            result = self._changelog_cache.get((serial, relpath), absent)
         if result is absent:
             changes = self._changelog_cache.get(serial, absent)
             if changes is not absent and relpath in changes:
                 (keyname, back_serial, value) = changes[relpath]
                 result = (serial, back_serial, value)
         if result is absent:
-            result = get_relpath_at(self, relpath, serial)
-        self._changelog_cache.put((serial, relpath), result)
+            result = self._get_relpath_at(relpath, serial)
+        if gettotalsizeof(result, maxlen=100000) is None:
+            # result is big, put it in the changelog cache,
+            # which has fewer entries to preserve memory
+            self._changelog_cache.put((serial, relpath), result)
+        else:
+            # result is small
+            self._relpath_cache.put((serial, relpath), result)
         return result
 
     def iter_relpaths_at(self, typedkeys, at_serial):
@@ -139,10 +171,7 @@ class BaseConnection:
                 **keyname_id_values))
             if not rows:
                 continue
-            changes = self._changelog_cache.get(serial, absent)
-            if changes is absent:
-                changes = loads(
-                    self.get_raw_changelog_entry(serial))[0]
+            changes = self.get_changes(serial)
             for relpath, keyname, serial in rows:
                 (keyname, back_serial, val) = changes[relpath]
                 yield RelpathInfo(
@@ -158,20 +187,15 @@ class Connection(BaseConnection):
 
     def io_file_exists(self, path):
         assert not os.path.isabs(path)
-        c = self._sqlconn.cursor()
         q = "SELECT path FROM files WHERE path = ?"
-        c.execute(q, (path,))
-        result = c.fetchone()
-        c.close()
+        result = self.fetchone(q, (path,))
         return result is not None
 
     def io_file_set(self, path, content):
         assert not os.path.isabs(path)
         assert not path.endswith("-tmp")
-        c = self._sqlconn.cursor()
         q = "INSERT OR REPLACE INTO files (path, size, data) VALUES (?, ?, ?)"
-        c.execute(q, (path, len(content), sqlite3.Binary(content)))
-        c.close()
+        self.fetchone(q, (path, len(content), sqlite3.Binary(content)))
         self.dirty_files[path] = True
 
     def io_file_open(self, path):
@@ -179,31 +203,23 @@ class Connection(BaseConnection):
 
     def io_file_get(self, path):
         assert not os.path.isabs(path)
-        c = self._sqlconn.cursor()
         q = "SELECT data FROM files WHERE path = ?"
-        c.execute(q, (path,))
-        content = c.fetchone()
-        c.close()
+        content = self.fetchone(q, (path,))
         if content is None:
             raise IOError()
         return bytes(content[0])
 
     def io_file_size(self, path):
         assert not os.path.isabs(path)
-        c = self._sqlconn.cursor()
         q = "SELECT size FROM files WHERE path = ?"
-        c.execute(q, (path,))
-        result = c.fetchone()
-        c.close()
+        result = self.fetchone(q, (path,))
         if result is not None:
             return result[0]
 
     def io_file_delete(self, path):
         assert not os.path.isabs(path)
-        c = self._sqlconn.cursor()
         q = "DELETE FROM files WHERE path = ?"
-        c.execute(q, (path,))
-        c.close()
+        self.fetchone(q, (path,))
         self.dirty_files[path] = None
 
     def write_transaction(self):
@@ -218,7 +234,10 @@ class BaseStorage(object):
         self.basedir = basedir
         self.sqlpath = self.basedir.join(self.db_filename)
         self._notify_on_commit = notify_on_commit
-        self._changelog_cache = LRUCache(cache_size)  # is thread safe
+        changelog_cache_size = max(1, cache_size // 20)
+        relpath_cache_size = max(1, cache_size - changelog_cache_size)
+        self._changelog_cache = LRUCache(changelog_cache_size)  # is thread safe
+        self._relpath_cache = LRUCache(relpath_cache_size)  # is thread safe
         self.last_commit_timestamp = time.time()
         self.ensure_tables_exist()
 
@@ -289,15 +308,18 @@ class BaseStorage(object):
         uri = "file:%s?mode=%s" % (self.sqlpath, mode)
         sqlconn = self._get_sqlconn(uri)
         if write:
-            start_time = time.time()
+            start_time = time.monotonic()
+            thread = current_thread()
             while 1:
                 try:
                     sqlconn.execute("begin immediate")
                     break
                 except sqlite3.OperationalError:
                     # another thread may be writing, give it a chance to finish
-                    time.sleep(0)
-                    if time.time() - start_time > 5:
+                    time.sleep(0.01)
+                    if hasattr(thread, "exit_if_shutdown"):
+                        thread.exit_if_shutdown()
+                    if time.monotonic() - start_time > 30:
                         # if it takes this long, something is wrong
                         raise
         conn = self.Connection(sqlconn, self.basedir, self)
@@ -336,6 +358,7 @@ class BaseStorage(object):
                     q = objs.pop(name)
                     c.execute(q)
                 assert not objs
+            c.close()
             conn.commit()
         assert not missing
 
@@ -389,24 +412,63 @@ def devpiserver_metrics(request):
     storage = xom.keyfs._storage
     if not isinstance(storage, BaseStorage):
         return result
-    cache = getattr(storage, '_changelog_cache', None)
-    if cache is None:
+    changelog_cache = getattr(storage, '_changelog_cache', None)
+    relpath_cache = getattr(storage, '_relpath_cache', None)
+    if changelog_cache is None and relpath_cache is None:
         return result
+    # get sizes for changelog_cache
+    evictions = changelog_cache.evictions if changelog_cache else 0
+    hits = changelog_cache.hits if changelog_cache else 0
+    lookups = changelog_cache.lookups if changelog_cache else 0
+    misses = changelog_cache.misses if changelog_cache else 0
+    size = changelog_cache.size if changelog_cache else 0
+    # add sizes for relpath_cache
+    evictions += relpath_cache.evictions if relpath_cache else 0
+    hits += relpath_cache.hits if relpath_cache else 0
+    lookups += relpath_cache.lookups if relpath_cache else 0
+    misses += relpath_cache.misses if relpath_cache else 0
+    size += relpath_cache.size if relpath_cache else 0
     result.extend([
-        ('devpi_server_storage_cache_evictions', 'counter', cache.evictions),
-        ('devpi_server_storage_cache_hits', 'counter', cache.hits),
-        ('devpi_server_storage_cache_lookups', 'counter', cache.lookups),
-        ('devpi_server_storage_cache_misses', 'counter', cache.misses),
-        ('devpi_server_storage_cache_size', 'gauge', cache.size)])
+        ('devpi_server_storage_cache_evictions', 'counter', evictions),
+        ('devpi_server_storage_cache_hits', 'counter', hits),
+        ('devpi_server_storage_cache_lookups', 'counter', lookups),
+        ('devpi_server_storage_cache_misses', 'counter', misses),
+        ('devpi_server_storage_cache_size', 'gauge', size)])
+    if changelog_cache:
+        result.extend([
+            ('devpi_server_changelog_cache_evictions', 'counter', changelog_cache.evictions),
+            ('devpi_server_changelog_cache_hits', 'counter', changelog_cache.hits),
+            ('devpi_server_changelog_cache_lookups', 'counter', changelog_cache.lookups),
+            ('devpi_server_changelog_cache_misses', 'counter', changelog_cache.misses),
+            ('devpi_server_changelog_cache_size', 'gauge', changelog_cache.size),
+            ('devpi_server_changelog_cache_items', 'gauge', len(changelog_cache.data) if changelog_cache.data else 0)])
+    if relpath_cache:
+        result.extend([
+            ('devpi_server_relpath_cache_evictions', 'counter', relpath_cache.evictions),
+            ('devpi_server_relpath_cache_hits', 'counter', relpath_cache.hits),
+            ('devpi_server_relpath_cache_lookups', 'counter', relpath_cache.lookups),
+            ('devpi_server_relpath_cache_misses', 'counter', relpath_cache.misses),
+            ('devpi_server_relpath_cache_size', 'gauge', relpath_cache.size),
+            ('devpi_server_relpath_cache_items', 'gauge', len(relpath_cache.data) if relpath_cache.data else 0)])
     return result
 
 
+class LazyChangesFormatter:
+    __slots__ = ('keys',)
+
+    def __init__(self, changes):
+        self.keys = changes.keys()
+
+    def __str__(self):
+        return f"keys: {','.join(repr(c) for c in self.keys)}"
+
+
 class Writer:
     def __init__(self, storage, conn):
         self.conn = conn
         self.storage = storage
         self.changes = {}
-        self.next_serial = conn.last_changelog_serial + 1
+        self.commit_serial = conn.last_changelog_serial + 1
 
     def record_set(self, typedkey, value=None, back_serial=None):
         """ record setting typedkey to value (None means it's deleted) """
@@ -416,27 +478,26 @@ class Writer:
                 _, back_serial = self.conn.db_read_typedkey(typedkey.relpath)
             except KeyError:
                 back_serial = -1
-        self.conn.db_write_typedkey(typedkey.relpath, typedkey.name, self.next_serial)
+        self.conn.db_write_typedkey(typedkey.relpath, typedkey.name, self.commit_serial)
         # at __exit__ time we write out changes to the _changelog_cache
         # so we protect here against the caller modifying the value later
         value = get_mutable_deepcopy(value)
         self.changes[typedkey.relpath] = (typedkey.name, back_serial, value)
 
     def __enter__(self):
-        self.log = thread_push_log("fswriter%s:" % self.next_serial)
+        self.log = thread_push_log("fswriter%s:" % self.commit_serial)
         return self
 
     def __exit__(self, cls, val, tb):
-        commit_serial = self.next_serial
+        commit_serial = self.commit_serial
         thread_pop_log("fswriter%s:" % commit_serial)
         if cls is None:
             entry = self.changes, []
             self.conn.write_changelog_entry(commit_serial, entry)
             self.conn.commit()
-            message = "committed: keys: %s"
-            args = [",".join(map(repr, list(self.changes)))]
             self.log.info("commited at %s", commit_serial)
-            self.log.debug(message, *args)
+            self.log.debug(
+                "committed: %s", LazyChangesFormatter(self.changes))
 
             self.storage._notify_on_commit(commit_serial)
         else:
