@@ -1,6 +1,5 @@
 from .config import hookimpl
-from .fileutil import BytesForHardlink
-from .interfaces import IStorageConnection2
+from .interfaces import IStorageConnection3
 from .keyfs_sqlite import BaseConnection
 from .keyfs_sqlite import BaseStorage
 from .log import threadlog, thread_push_log, thread_pop_log
@@ -8,17 +7,24 @@ from .readonly import ReadonlyView
 from .readonly import get_mutable_deepcopy
 from .fileutil import get_write_file_ensure_dir, rename, loads
 from hashlib import sha256
+from zope.interface import Interface
+from zope.interface import alsoProvides
 from zope.interface import implementer
 import errno
 import os
 import re
+import shutil
 import sys
 import threading
 import time
 
 
+class IStorageFile(Interface):
+    """ Marker interface. """
+
+
 class DirtyFile(object):
-    def __init__(self, path, content):
+    def __init__(self, path):
         self.path = path
         # use hash of path, pid and thread id to prevent conflicts
         key = "%s%i%i" % (
@@ -28,8 +34,12 @@ class DirtyFile(object):
             # on windows we have to shorten the digest, otherwise we reach
             # the 260 chars file path limit too quickly
             digest = digest[:8]
-        self.tmppath = '%s-%s-tmp' % (path, digest)
-        if isinstance(content, BytesForHardlink):
+        self.tmppath = f"{path}-{digest}-tmp"
+
+    @classmethod
+    def from_content(cls, path, content_or_file):
+        self = DirtyFile(path)
+        if hasattr(content_or_file, "devpi_srcpath"):
             dirname = os.path.dirname(self.tmppath)
             if not os.path.exists(dirname):
                 try:
@@ -40,13 +50,24 @@ class DirtyFile(object):
                     # another thread tries to create the same folder
                     if e.errno != errno.EEXIST:
                         raise
-            os.link(content.devpi_srcpath, self.tmppath)
+            os.link(content_or_file.devpi_srcpath, self.tmppath)
         else:
             with get_write_file_ensure_dir(self.tmppath) as f:
-                f.write(content)
+                if not isinstance(content_or_file, bytes) and not callable(getattr(content_or_file, "seekable", None)):
+                    content_or_file = content_or_file.read()
+                    if len(content_or_file) > 1048576:
+                        threadlog.warn(
+                            "Read %.1f megabytes into memory in keyfs_sqlite_fs from_content for %s, because of unseekable file",
+                            len(content_or_file) / 1048576, path)
+                if isinstance(content_or_file, bytes):
+                    f.write(content_or_file)
+                else:
+                    content_or_file.seek(0)
+                    shutil.copyfileobj(content_or_file, f)
+        return self
 
 
-@implementer(IStorageConnection2)
+@implementer(IStorageConnection3)
 class Connection(BaseConnection):
     def rollback(self):
         BaseConnection.rollback(self)
@@ -67,10 +88,21 @@ class Connection(BaseConnection):
             path = dirty_file.tmppath
         return os.path.exists(path)
 
-    def io_file_set(self, path, content):
+    def io_file_set(self, path, content_or_file):
         path = self._basedir.join(path).strpath
         assert not path.endswith("-tmp")
-        self.dirty_files[path] = DirtyFile(path, content)
+        if IStorageFile.providedBy(content_or_file):
+            self.dirty_files[path] = DirtyFile(path)
+        else:
+            self.dirty_files[path] = DirtyFile.from_content(path, content_or_file)
+
+    def io_file_new_open(self, path):
+        path = self._basedir.join(path).strpath
+        assert not path.endswith("-tmp")
+        assert not self.io_file_exists(path)
+        f = get_write_file_ensure_dir(DirtyFile(path).tmppath)
+        alsoProvides(f, IStorageFile)
+        return f
 
     def io_file_open(self, path):
         path = self._basedir.join(path).strpath
@@ -89,7 +121,12 @@ class Connection(BaseConnection):
                 raise IOError()
             path = dirty_file.tmppath
         with open(path, "rb") as f:
-            return f.read()
+            data = f.read()
+            if len(data) > 1048576:
+                threadlog.warn(
+                    "Read %.1f megabytes into memory in io_file_get for %s",
+                    len(data) / 1048576, path)
+            return data
 
     def io_file_size(self, path):
         path = self._basedir.join(path).strpath
@@ -118,16 +155,10 @@ class Connection(BaseConnection):
         basedir = str(self.storage.basedir)
         rel_renames = list(make_rel_renames(basedir, pending_renames))
         files_commit, files_del = commit_renames(basedir, rel_renames)
-        message = "wrote files without increasing serial: "
-        args = []
-        if files_commit:
-            message += "%s"
-            args.append(",".join(files_commit))
-        if files_del:
-            message += ", files_del: %s"
-            args.append(",".join(files_del))
-        if args:
-            threadlog.debug(message, *args)
+        if files_commit or files_del:
+            threadlog.debug(
+                "wrote files without increasing serial: %s",
+                LazyChangesFormatter({}, files_commit, files_del))
 
 
 class Storage(BaseStorage):
@@ -172,12 +203,31 @@ def devpiserver_storage_backend(settings):
         _test_markers=["storage_with_filesystem"])
 
 
+class LazyChangesFormatter:
+    __slots__ = ('files_commit', 'files_del', 'keys')
+
+    def __init__(self, changes, files_commit, files_del):
+        self.files_commit = files_commit
+        self.files_del = files_del
+        self.keys = changes.keys()
+
+    def __str__(self):
+        msg = []
+        if self.keys:
+            msg.append(f"keys: {','.join(repr(c) for c in self.keys)}")
+        if self.files_commit:
+            msg.append(f"files_commit: {','.join(self.files_commit)}")
+        if self.files_del:
+            msg.append(f"files_del: {','.join(self.files_del)}")
+        return ", ".join(msg)
+
+
 class FSWriter:
     def __init__(self, storage, conn):
         self.conn = conn
         self.storage = storage
         self.changes = {}
-        self.next_serial = conn.last_changelog_serial + 1
+        self.commit_serial = conn.last_changelog_serial + 1
 
     def record_set(self, typedkey, value=None, back_serial=None):
         """ record setting typedkey to value (None means it's deleted) """
@@ -187,36 +237,26 @@ class FSWriter:
                 _, back_serial = self.conn.db_read_typedkey(typedkey.relpath)
             except KeyError:
                 back_serial = -1
-        self.conn.db_write_typedkey(typedkey.relpath, typedkey.name, self.next_serial)
+        self.conn.db_write_typedkey(typedkey.relpath, typedkey.name, self.commit_serial)
         # at __exit__ time we write out changes to the _changelog_cache
         # so we protect here against the caller modifying the value later
         value = get_mutable_deepcopy(value)
         self.changes[typedkey.relpath] = (typedkey.name, back_serial, value)
 
     def __enter__(self):
-        self.log = thread_push_log("fswriter%s:" % self.next_serial)
+        self.log = thread_push_log("fswriter%s:" % self.commit_serial)
         return self
 
     def __exit__(self, cls, val, tb):
-        commit_serial = self.next_serial
+        commit_serial = self.commit_serial
         thread_pop_log("fswriter%s:" % commit_serial)
         if cls is None:
             pending_renames = write_dirty_files(self.conn.dirty_files)
 
-            changed_keys, files_commit, files_del = \
-                self.commit_to_filesystem(pending_renames)
+            changes_formatter = self.commit_to_filesystem(pending_renames)
 
-            # write out a nice commit entry to logging
-            message = "committed: keys: %s"
-            args = [",".join(map(repr, changed_keys))]
-            if files_commit:
-                message += ", files_commit: %s"
-                args.append(",".join(files_commit))
-            if files_del:
-                message += ", files_del: %s"
-                args.append(",".join(files_del))
             self.log.info("commited at %s", commit_serial)
-            self.log.debug(message, *args)
+            self.log.debug("committed: keys: %s", changes_formatter)
 
             self.storage._notify_on_commit(commit_serial)
         else:
@@ -230,7 +270,7 @@ class FSWriter:
             make_rel_renames(basedir, pending_renames)
         )
         entry = self.changes, rel_renames
-        self.conn.write_changelog_entry(self.next_serial, entry)
+        self.conn.write_changelog_entry(self.commit_serial, entry)
         self.conn.commit()
 
         # If we crash in the remainder, the next restart will
@@ -239,7 +279,7 @@ class FSWriter:
         # - initialize next_serial from the max committed serial + 1
         files_commit, files_del = commit_renames(basedir, rel_renames)
         self.storage.last_commit_timestamp = time.time()
-        return list(self.changes), files_commit, files_del
+        return LazyChangesFormatter(self.changes, files_commit, files_del)
 
 
 def drop_dirty_files(dirty_files):
