@@ -10,6 +10,7 @@ import os
 import asyncio
 import py
 import sys
+import time
 import traceback
 
 from requests import Response, exceptions
@@ -196,8 +197,15 @@ class AsyncioLoopThread(object):
                 return
 
     def thread_shutdown(self):
-        self.loop.call_soon_threadsafe(self.loop.stop)
-        self.thread.join()
+        loop = self.loop
+        try:
+            loop.call_soon_threadsafe(loop.stop)
+        except RuntimeError:
+            # the shutdown function may have been called already
+            return
+        while loop.is_running():
+            time.sleep(0.1)
+        loop.close()
 
 
 class XOM:
@@ -295,11 +303,10 @@ class XOM:
                 return res
 
         app = xom.create_app()
-        with xom.thread_pool.live():
-            if xom.is_replica():
-                # XXX ground restart_as_write_transaction better
-                xom.keyfs.restart_as_write_transaction = None
-            return wsgi_run(xom, app)
+        if xom.is_replica():
+            # XXX ground restart_as_write_transaction better
+            xom.keyfs.restart_as_write_transaction = None
+        return xom.thread_pool.run(wsgi_run, xom, app)
 
     def fatal(self, msg):
         self.keyfs.release_all_wait_tx()
@@ -340,6 +347,9 @@ class XOM:
         max_retries = self.config.replica_max_retries
         return self.new_http_session("server", max_retries=max_retries)
 
+    def _close_sessions(self):
+        self._httpsession.close()
+
     async def async_httpget(self, url, allow_redirects, timeout=None, extra_headers=None):
         timeout = aiohttp.ClientTimeout(total=timeout)
         async with aiohttp.ClientSession(timeout=timeout, trust_env=True) as session:
@@ -407,14 +417,15 @@ class XOM:
     view_deriver.options = ('is_mutating',)
 
     def create_app(self):
+        from devpi_server.middleware import OutsideURLMiddleware
         from devpi_server.view_auth import DevpiSecurityPolicy
         from devpi_server.views import ContentTypePredicate
-        from devpi_server.views import OutsideURLMiddleware
         from devpi_server.views import route_url, INSTALLER_USER_AGENT
         from pkg_resources import get_distribution
         from pyramid.config import Configurator
         from pyramid.viewderivers import INGRESS
         log = self.log
+        log.info("running with role %r", self.config.role)
         log.debug("creating application in process %s", os.getpid())
         pyramid_config = Configurator(root_factory='devpi_server.view_auth.RootFactory')
         pyramid_config.set_security_policy(DevpiSecurityPolicy(self))
@@ -453,36 +464,64 @@ class XOM:
 
         pyramid_config.add_route("/+status", "/+status")
         pyramid_config.add_route("/+api", "/+api", accept="application/json")
-        pyramid_config.add_route("/{user}/+api", "/{user}/+api", accept="application/json")
-        pyramid_config.add_route("/{user}/{index}/+api", "/{user}/{index}/+api", accept="application/json")
+        pyramid_config.add_route(
+            "/{user}/+api",
+            "/{user:[^+/]+}/+api",
+            accept="application/json")
+        pyramid_config.add_route(
+            "/{user}/{index}/+api",
+            "/{user:[^+/]+}/{index:[^+/]+}/+api",
+            accept="application/json")
         pyramid_config.add_route("/+authcheck", "/+authcheck")
         pyramid_config.add_route("/+login", "/+login", accept="application/json")
-        pyramid_config.add_route("/{user}/{index}/+e/{relpath:.*}", "/{user}/{index}/+e/{relpath:.*}")
-        pyramid_config.add_route("/{user}/{index}/+f/{relpath:.*}", "/{user}/{index}/+f/{relpath:.*}")
-        pyramid_config.add_route("/{user}/{index}/+simple", "/{user}/{index}/+simple")
-        pyramid_config.add_route("/{user}/{index}/+simple/", "/{user}/{index}/+simple/")
-        pyramid_config.add_route("/{user}/{index}/+simple/{project}",
-                                 "/{user}/{index}/+simple/{project}")
-        pyramid_config.add_route("/{user}/{index}/+simple/{project}/",
-                                 "/{user}/{index}/+simple/{project}/")
-        pyramid_config.add_route("/{user}/{index}/+simple/{project}/refresh",
-                                 "/{user}/{index}/+simple/{project}/refresh")
-        pyramid_config.add_route("/{user}/{index}/{project}/{version}",
-                                 "/{user}/{index}/{project}/{version:[^/]+/?}")
         pyramid_config.add_route(
-            "installer_simple", "/{user}/{index:[^/]+/?}",
+            "/{user}/{index}/+e/{relpath:.*}",
+            "/{user:[^+/]+}/{index:[^+/]+}/+e/{relpath:.*}")
+        pyramid_config.add_route(
+            "/{user}/{index}/+f/{relpath:.*}",
+            "/{user:[^+/]+}/{index:[^+/]+}/+f/{relpath:.*}")
+        pyramid_config.add_route(
+            "/{user}/{index}/+simple",
+            "/{user:[^+/]+}/{index:[^+/]+}/+simple")
+        pyramid_config.add_route(
+            "/{user}/{index}/+simple/",
+            "/{user:[^+/]+}/{index:[^+/]+}/+simple/")
+        pyramid_config.add_route(
+            "/{user}/{index}/+simple/{project}",
+            "/{user:[^+/]+}/{index:[^+/]+}/+simple/{project:[^+/]+}")
+        pyramid_config.add_route(
+            "/{user}/{index}/+simple/{project}/",
+            "/{user:[^+/]+}/{index:[^+/]+}/+simple/{project:[^+/]+}/")
+        pyramid_config.add_route(
+            "/{user}/{index}/+simple/{project}/refresh",
+            "/{user:[^+/]+}/{index:[^+/]+}/+simple/{project:[^+/]+}/refresh")
+        pyramid_config.add_route(
+            "/{user}/{index}/{project}/{version}",
+            "/{user:[^+/]+}/{index:[^+/]+}/{project:[^+/]+}/{version:[^/]+/?}")
+        pyramid_config.add_route(
+            "installer_simple",
+            "/{user:[^+/]+}/{index:[^+/]+/?}",
             header="User-Agent:" + INSTALLER_USER_AGENT,
             accept="text/html", request_method="GET")
         pyramid_config.add_route(
-            "installer_simple_project", "/{user}/{index}/{project:[^/]+/?}",
+            "installer_simple_project", "/{user:[^+/]+}/{index}/{project:[^+/]+/?}",
             header="User-Agent:" + INSTALLER_USER_AGENT,
             accept="text/html", request_method="GET")
-        pyramid_config.add_route("/{user}/{index}/{project}",
-                                 "/{user}/{index}/{project:[^/]+/?}")
-        pyramid_config.add_route("/{user}/{index}/", "/{user}/{index}/")
-        pyramid_config.add_route("/{user}/{index}", "/{user}/{index}")
-        pyramid_config.add_route("/{user}/", "/{user}/")
-        pyramid_config.add_route("/{user}", "/{user}")
+        pyramid_config.add_route(
+            "/{user}/{index}/{project}",
+            "/{user:[^+/]+}/{index:[^+/]+}/{project:[^+/]+/?}")
+        pyramid_config.add_route(
+            "/{user}/{index}/",
+            "/{user:[^+/]+}/{index:[^+/]+}/")
+        pyramid_config.add_route(
+            "/{user}/{index}",
+            "/{user:[^+/]+}/{index:[^+/]+}")
+        pyramid_config.add_route(
+            "/{user}/",
+            "/{user:[^+/]+}/")
+        pyramid_config.add_route(
+            "/{user}",
+            "/{user:[^+/]+}")
         pyramid_config.add_route("/", "/")
 
         # register tweens for logging, transaction and replication
@@ -522,6 +561,9 @@ class FatalResponse:
     def __repr__(self):
         return "%s(%r)" % (self.__class__.__name__, self.reason)
 
+    def close(self):
+        pass
+
 
 def get_remote_ip(request):
     return request.headers.get("X-REAL-IP", request.client_addr)
