@@ -23,6 +23,7 @@ from time import gmtime, strftime
 from .auth import hash_password, verify_and_update_password_hash
 from .config import hookimpl
 from .filestore import FileEntry
+from .filestore import get_file_hash
 from .log import threadlog
 from .readonly import get_mutable_deepcopy
 
@@ -742,19 +743,20 @@ class BaseStage(object):
     def store_toxresult(self, link, toxresultdata):
         if self.customizer.readonly:
             raise ReadonlyIndex("index is marked read only")
-        assert isinstance(toxresultdata, dict), toxresultdata
         linkstore = self.get_linkstore_perstage(link.project, link.version, readonly=False)
+        if isinstance(toxresultdata, dict):
+            toxresultdata = json.dumps(toxresultdata).encode("utf-8")
         return linkstore.new_reflink(
-                rel="toxresult",
-                file_content=json.dumps(toxresultdata).encode("utf-8"),
-                for_entrypath=link)
+            rel="toxresult",
+            content_or_file=toxresultdata,
+            for_entrypath=link)
 
     def get_toxresults(self, link):
         l = []
         linkstore = self.get_linkstore_perstage(link.project, link.version)
         for reflink in linkstore.get_links(rel="toxresult", for_entrypath=link):
-            data = reflink.entry.file_get_content().decode("utf-8")
-            l.append(json.loads(data))
+            with reflink.entry.file_open_read() as f:
+                l.append(json.load(f))
         return l
 
     def filter_versions(self, project, versions):
@@ -1274,7 +1276,7 @@ class PrivateStage(BaseStage):
     def has_project_perstage(self, project):
         return normalize_name(project) in self.list_projects_perstage()
 
-    def store_releasefile(self, project, version, filename, content,
+    def store_releasefile(self, project, version, filename, content_or_file,
                           last_modified=None):
         if self.customizer.readonly:
             raise ReadonlyIndex("index is marked read only")
@@ -1291,14 +1293,14 @@ class PrivateStage(BaseStage):
                 raise MissesRegistration("%s-%s", project, version)
         linkstore = self.get_linkstore_perstage(project, version, readonly=False)
         link = linkstore.create_linked_entry(
-                rel="releasefile",
-                basename=filename,
-                file_content=content,
-                last_modified=last_modified)
+            rel="releasefile",
+            basename=filename,
+            content_or_file=content_or_file,
+            last_modified=last_modified)
         self._regen_simplelinks(project)
         return link
 
-    def store_doczip(self, project, version, content):
+    def store_doczip(self, project, version, content_or_file):
         if self.customizer.readonly:
             raise ReadonlyIndex("index is marked read only")
         project = normalize_name(project)
@@ -1317,10 +1319,9 @@ class PrivateStage(BaseStage):
             self.set_versiondata({'name': project, 'version': version})
         linkstore = self.get_linkstore_perstage(project, version, readonly=False)
         link = linkstore.create_linked_entry(
-                rel="doczip",
-                basename=basename,
-                file_content=content,
-        )
+            rel="doczip",
+            basename=basename,
+            content_or_file=content_or_file)
         return link
 
     def get_doczip_entry(self, project, version):
@@ -1436,11 +1437,11 @@ class ELink(object):
     def hash_type(self):
         return self.hash_spec.split("=")[0]
 
-    def matches_checksum(self, content):
+    def matches_checksum(self, content_or_file):
         hash_algo, hash_value = parse_hash_spec(self.hash_spec)
         if not hash_algo:
             return True
-        return hash_algo(content).hexdigest() == hash_value
+        return get_file_hash(content_or_file, hash_algo().name) == hash_value
 
     def __getattr__(self, name):
         try:
@@ -1479,8 +1480,9 @@ class LinkStore:
         self.version = version
         self.verdata = stage.get_versiondata_perstage(self.project, version, readonly=readonly)
         if not self.verdata:
-            raise MissesRegistration("%s-%s on stage %s",
-                                     project, version, stage.name)
+            raise MissesRegistration(
+                "%s-%s on stage %s at %s",
+                project, version, stage.name, stage.keyfs.tx.at_serial)
 
     @property
     def metadata(self):
@@ -1493,8 +1495,7 @@ class LinkStore:
     def get_file_entry(self, relpath):
         return self.filestore.get_file_entry(relpath)
 
-    def create_linked_entry(self, rel, basename, file_content, last_modified=None):
-        assert isinstance(file_content, bytes)
+    def create_linked_entry(self, rel, basename, content_or_file, last_modified=None):
         overwrite = None
         for link in self.get_links(rel=rel, basename=basename):
             if not self.stage.ixconfig.get("volatile"):
@@ -1506,7 +1507,7 @@ class LinkStore:
             overwrite = sum(x.get('count', 0)
                             for x in link.get_logs() if x.get('what') == 'overwrite')
             self.remove_links(rel=rel, basename=basename)
-        file_entry = self._create_file_entry(basename, file_content)
+        file_entry = self._create_file_entry(basename, content_or_file)
         if last_modified is not None:
             file_entry.last_modified = last_modified
         link = self._add_link_to_file_entry(rel, file_entry)
@@ -1514,19 +1515,20 @@ class LinkStore:
             link.add_log('overwrite', None, count=overwrite + 1)
         return link
 
-    def new_reflink(self, rel, file_content, for_entrypath):
+    def new_reflink(self, rel, content_or_file, for_entrypath):
         if isinstance(for_entrypath, ELink):
             for_entrypath = for_entrypath.entrypath
         links = self.get_links(entrypath=for_entrypath)
-        assert len(links) == 1, "need exactly one reference, got %s" %(links,)
+        assert len(links) == 1, f"need exactly one reference, got {links}"
         base_entry = links[0].entry
         other_reflinks = self.get_links(rel=rel, for_entrypath=for_entrypath)
         timestamp = strftime("%Y%m%d%H%M%S", gmtime())
         filename = "%s.%s-%s-%d" % (
             base_entry.basename, rel, timestamp, len(other_reflinks))
-        entry = self._create_file_entry(filename, file_content,
-                                        ref_hash_spec=base_entry.hash_spec)
-        return self._add_link_to_file_entry(rel, entry, for_entrypath=for_entrypath)
+        entry = self._create_file_entry(
+            filename, content_or_file, ref_hash_spec=base_entry.hash_spec)
+        return self._add_link_to_file_entry(
+            rel, entry, for_entrypath=for_entrypath)
 
     def remove_links(self, rel=None, basename=None, for_entrypath=None):
         linkdicts = self._get_inplace_linkdicts()
@@ -1556,12 +1558,12 @@ class LinkStore:
         return list(filter(fil, [ELink(self.filestore, linkdict, self.project, self.version)
                            for linkdict in self.verdata.get("+elinks", [])]))
 
-    def _create_file_entry(self, basename, file_content, ref_hash_spec=None):
+    def _create_file_entry(self, basename, content_or_file, ref_hash_spec=None):
         entry = self.filestore.store(
-                    user=self.stage.username, index=self.stage.index,
-                    basename=basename,
-                    file_content=file_content,
-                    dir_hash_spec=ref_hash_spec)
+            user=self.stage.username, index=self.stage.index,
+            basename=basename,
+            content_or_file=content_or_file,
+            dir_hash_spec=ref_hash_spec)
         entry.project = self.project
         entry.version = self.version
         return entry
@@ -1616,10 +1618,10 @@ def normalize_bases(model, bases):
         try:
             stage_base = model.getstage(base)
         except ValueError:
-            messages.append("invalid base index spec: %r" % (base,))
+            messages.append(f"invalid base index spec: {base!r}")
         else:
             if stage_base is None:
-                messages.append("base index %r does not exist" %(base,))
+                messages.append(f"base index {base!r} does not exist")
             else:
                 newbases.append(stage_base.name)
     if messages:
