@@ -1,20 +1,15 @@
 from __future__ import unicode_literals
-
+import contextlib
 import os
 import py
 import re
 import traceback
 from time import time
-try:
-    from collections.abc import Iterator
-except ImportError:
-    from collections import Iterator
 from devpi_common.types import ensure_unicode
 from devpi_common.url import URL
 from devpi_common.metadata import get_pyversion_filetype
 import devpi_server
 from html import escape
-from io import BytesIO
 from lazy import lazy
 from pluggy import HookimplMarker
 from pyramid.authentication import b64encode
@@ -28,6 +23,7 @@ from pyramid.httpexceptions import HTTPUnauthorized
 from pyramid.httpexceptions import exception_response
 from pyramid.request import Request
 from pyramid.request import apply_request_extensions
+from pyramid.response import FileIter
 from pyramid.response import Response
 from pyramid.security import forget
 from pyramid.threadlocal import RequestContext
@@ -42,6 +38,8 @@ from devpi_common.validation import normalize_name, is_valid_archive_name
 
 from .config import hookimpl
 from .filestore import BadGateway
+from .filestore import get_checksum_error
+from .fileutil import buffered_iterator
 from .model import InvalidIndex, InvalidIndexconfig, InvalidUser, InvalidUserconfig
 from .model import ReadonlyIndex
 from .model import RemoveValue
@@ -146,26 +144,6 @@ class ContentTypePredicate(object):
         return request.content_type == self.val
 
 
-class OutsideURLMiddleware(object):
-    def __init__(self, app, xom):
-        self.app = app
-        self.xom = xom
-
-    def __call__(self, environ, start_response):
-        outside_url = environ.get('HTTP_X_OUTSIDE_URL')
-        if not outside_url:
-            outside_url = self.xom.config.args.outside_url
-        if outside_url:
-            # XXX memoize it for later access from replica thread
-            # self.xom.current_outside_url = outside_url
-            outside_url = urlparse(outside_url)
-            environ['wsgi.url_scheme'] = outside_url.scheme
-            environ['HTTP_HOST'] = outside_url.netloc
-            if outside_url.path:
-                environ['SCRIPT_NAME'] = outside_url.path
-        return self.app(environ, start_response)
-
-
 def route_url(self, *args, **kw):
     url = super(self.__class__, self).route_url(*args, **kw)
     # Unquote plus signs in path segment. The settings in pyramid for
@@ -181,7 +159,7 @@ def tween_request_logging(handler, registry):
     nodeinfo = registry["xom"].config.nodeinfo
 
     def request_log_handler(request):
-        tag = "[req%s]" %(next(req_count))
+        tag = f"[req{next(req_count)}]"
         log = thread_push_log(tag)
         try:
             request.log = log
@@ -230,8 +208,6 @@ def tween_keyfs_transaction(handler, registry):
 
 
 def set_header_devpi_serial(response, tx):
-    if isinstance(response._app_iter, Iterator):
-        return
     if tx.commit_serial is not None:
         serial = tx.commit_serial
     else:
@@ -366,9 +342,9 @@ def devpiweb_get_status_info(request):
             pass
         elif sync_at is None and last_processed is None and (now - status["last-commit-timestamp"]) > 300:
             msgs.append(dict(status="fatal", msg="The event processing doesn't seem to start"))
-        elif sync_at is None or (sync_at > 21600):
+        elif (sync_at is None and last_processed is None) or (sync_at is not None and sync_at > 21600):
             msgs.append(dict(status="fatal", msg="The event processing hasn't been in sync for more than 6 hours"))
-        elif sync_at > 3600:
+        elif sync_at is not None and sync_at > 3600:
             msgs.append(dict(status="warn", msg="The event processing hasn't been in sync for more than 1 hour"))
         if sync_at is not None and (last_processed is None or (last_processed > 1800)):
             msgs.append(dict(status="fatal", msg="No changes processed by plugins for more than 30 minutes"))
@@ -490,6 +466,8 @@ class PyPIView:
         info = routes_mapper(orig_request)
         (orig_request.matchdict, orig_request.matched_route) = (
             info['match'], info['route'])
+        if orig_request.matched_route is None:
+            return HTTPForbidden()
         root_factory = orig_request.matched_route.factory or root_factory
         orig_request.context = root_factory(orig_request)
         with RequestContext(orig_request):
@@ -499,7 +477,8 @@ class PyPIView:
     # attach test results to release files
     #
 
-    @view_config(route_name="/{user}/{index}/+f/{relpath:.*}",
+    @view_config(
+        route_name="/{user}/{index}/+f/{relpath:.*}",
         request_method="POST",
         permission="toxresult_upload")
     def post_toxresult(self):
@@ -575,11 +554,9 @@ class PyPIView:
             whitelist_info = stage.get_mirror_whitelist_info(project)
             embed_form = whitelist_info['has_mirror_base']
             blocked_index = whitelist_info['blocked_by_mirror_whitelist']
-        body = b"".join(self._simple_list_project(
+        app_iter = buffered_iterator(self._simple_list_project(
             stage, project, result, embed_form, blocked_index))
-        response = Response(
-            content_length=len(body),
-            body=body)
+        response = Response(app_iter=app_iter)
         if stage.ixconfig['type'] == 'mirror':
             serial = stage.key_projsimplelinks(project).get().get("serial")
             if serial > 0:
@@ -673,12 +650,10 @@ class PyPIView:
         requested_by_installer = INSTALLER_USER_AGENT_REGEXP.match(
             self.request.user_agent or "")
         if requested_by_installer:
-            body = b"".join(self._simple_list_all_installer(stage, stage_results))
+            app_iter = buffered_iterator(self._simple_list_all_installer(stage, stage_results))
         else:
-            body = b"".join(self._simple_list_all(stage, stage_results))
-        return Response(
-            content_length=len(body),
-            body=body)
+            app_iter = buffered_iterator(self._simple_list_all(stage, stage_results))
+        return Response(app_iter=app_iter)
 
     def _simple_list_all(self, stage, stage_results):
         title = "%s: simple list (including inherited indices)" % (stage.name)
@@ -858,15 +833,16 @@ class PyPIView:
         try:
             linkstore = stage.get_linkstore_perstage(name, version)
         except stage.MissesRegistration:
-            apireturn(400, "there are no files for %s-%s on stage %s" %(
-                           name, version, stage.name))
-        links = dict([(rel, linkstore.get_links(rel=rel))
-                        for rel in ('releasefile', 'doczip', 'toxresult')])
+            apireturn(400, "there are no files for %s-%s on stage %s" % (
+                name, version, stage.name))
+        links = dict([
+            (rel, linkstore.get_links(rel=rel))
+            for rel in ('releasefile', 'doczip', 'toxresult')])
         if not links["releasefile"]:
             self.log.info("%s: no release files for version %s-%s" %
                           (stage.name, name, version))
-            apireturn(404, message="no release/files found for %s-%s" %(
-                      name, version))
+            apireturn(404, message="no release/files found for %s-%s" % (
+                name, version))
 
         if not request.has_permission("pkg_read"):
             abort(request, 403, "package read forbidden")
@@ -884,8 +860,8 @@ class PyPIView:
             self.log.debug("targetindex %r, auth_user %r", targetindex,
                            auth_user)
             if not request.has_permission("upload", context=target_stage):
-                apireturn(401, message="user %r cannot upload to %r"
-                                      %(auth_user, targetindex))
+                apireturn(401, message="user %r cannot upload to %r" % (
+                    auth_user, targetindex))
             self._set_versiondata_dict(target_stage, metadata)
             results.append((200, "register", name, version,
                             "->", target_stage.name))
@@ -907,68 +883,73 @@ class PyPIView:
             metadata["metadata_version"] = "2.1"
             self.log.info("registering %s-%s to %s", name, version, posturl)
             session = new_requests_session(agent=("server", server_version))
-            try:
-                r = session.post(posturl, data=metadata, auth=pypiauth)
-            except Exception as e:
-                exc_msg = ''.join(traceback.format_exception_only(e.__class__, e))
-                results.append((-1, "exception on register:", exc_msg))
-                apireturn(502, result=results, type="actionlog")
-            self.log.debug("register returned: %s", r.status_code)
-            results.append((r.status_code, "register", name, version))
-            ok_codes = (200, 201, 410)
-            proceed = (r.status_code in ok_codes)
-            if proceed:
-                for link in links["releasefile"]:
-                    entry = link.entry
-                    file_metadata = metadata.copy()
-                    file_metadata[":action"] = "file_upload"
-                    basename = link.basename
-                    pyver, filetype = get_pyversion_filetype(basename)
-                    file_metadata["filetype"] = filetype
-                    file_metadata["pyversion"] = pyver
-                    file_metadata["%s_digest" % link.hash_type] = link.hash_value
-                    content = entry.file_get_content()
-                    self.log.info("sending %s to %s, metadata %s",
-                             basename, posturl, file_metadata)
-                    try:
-                        r = session.post(
-                            posturl, data=file_metadata, auth=pypiauth,
-                            files={"content": (basename, content)})
-                    except Exception as e:
-                        exc_msg = ''.join(traceback.format_exception_only(e.__class__, e))
-                        results.append((-1, "exception on release upload:", exc_msg))
-                    else:
-                        self.log.debug("send finished, status: %s", r.status_code)
-                        results.append((r.status_code, "upload", entry.relpath,
-                                        r.text))
-                if links["doczip"]:
-                    doc_metadata = metadata.copy()
-                    doc_metadata[":action"] = "doc_upload"
-                    doczip = links["doczip"][0].entry.file_get_content()
-                    try:
-                        r = session.post(
-                            posturl, data=doc_metadata, auth=pypiauth,
-                            files={"content": (name + ".zip", doczip)})
-                    except Exception as e:
-                        exc_msg = ''.join(traceback.format_exception_only(e.__class__, e))
-                        results.append((-1, "exception on documentation upload:", exc_msg))
-                    else:
-                        self.log.debug("send finished, status: %s", r.status_code)
-                        results.append((r.status_code, "docfile", name))
-                #
-            if r.status_code in ok_codes:
-                apireturn(200, result=results, type="actionlog")
-            else:
-                apireturn(502, result=results, type="actionlog")
+            with contextlib.closing(session):
+                try:
+                    r = session.post(posturl, data=metadata, auth=pypiauth)
+                    r.close()
+                except Exception as e:
+                    exc_msg = ''.join(traceback.format_exception_only(e.__class__, e))
+                    results.append((-1, "exception on register:", exc_msg))
+                    apireturn(502, result=results, type="actionlog")
+                self.log.debug("register returned: %s", r.status_code)
+                results.append((r.status_code, "register", name, version))
+                ok_codes = (200, 201, 410)
+                proceed = (r.status_code in ok_codes)
+                if proceed:
+                    for link in links["releasefile"]:
+                        entry = link.entry
+                        file_metadata = metadata.copy()
+                        file_metadata[":action"] = "file_upload"
+                        basename = link.basename
+                        pyver, filetype = get_pyversion_filetype(basename)
+                        file_metadata["filetype"] = filetype
+                        file_metadata["pyversion"] = pyver
+                        file_metadata["%s_digest" % link.hash_type] = link.hash_value
+                        content = entry.file_get_content()
+                        self.log.info(
+                            "sending %s to %s, metadata %s",
+                            basename, posturl, file_metadata)
+                        try:
+                            r = session.post(
+                                posturl, data=file_metadata, auth=pypiauth,
+                                files={"content": (basename, content)})
+                        except Exception as e:
+                            exc_msg = ''.join(traceback.format_exception_only(e.__class__, e))
+                            results.append((-1, "exception on release upload:", exc_msg))
+                        else:
+                            self.log.debug("send finished, status: %s", r.status_code)
+                            results.append((r.status_code, "upload", entry.relpath,
+                                            r.text))
+                            r.close()
+                    if links["doczip"]:
+                        doc_metadata = metadata.copy()
+                        doc_metadata[":action"] = "doc_upload"
+                        doczip = links["doczip"][0].entry.file_get_content()
+                        try:
+                            r = session.post(
+                                posturl, data=doc_metadata, auth=pypiauth,
+                                files={"content": (name + ".zip", doczip)})
+                            r.close()
+                        except Exception as e:
+                            exc_msg = ''.join(traceback.format_exception_only(e.__class__, e))
+                            results.append((-1, "exception on documentation upload:", exc_msg))
+                        else:
+                            self.log.debug("send finished, status: %s", r.status_code)
+                            results.append((r.status_code, "docfile", name))
+                if r.status_code in ok_codes:
+                    apireturn(200, result=results, type="actionlog")
+                else:
+                    apireturn(502, result=results, type="actionlog")
 
     def _push_links(self, links, target_stage, name, version):
         for link in links["releasefile"]:
             if should_fetch_remote_file(link.entry, self.request.headers):
-                for part in iter_fetch_remote_file(self.xom, link.entry):
+                for part in iter_fetch_remote_file(self.xom, link.entry, link.entry.url):
                     pass
-            new_link = target_stage.store_releasefile(
-                name, version, link.basename, link.entry.file_get_content(),
-                last_modified=link.entry.last_modified)
+            with link.entry.file_open_read() as f:
+                new_link = target_stage.store_releasefile(
+                    name, version, link.basename, f,
+                    last_modified=link.entry.last_modified)
             new_link.add_logs(
                 x for x in link.get_logs()
                 if x.get('what') != 'overwrite')
@@ -984,9 +965,8 @@ class PyPIView:
             for toxlink in links["toxresult"]:
                 if toxlink.for_entrypath == link.entry.relpath:
                     ref_link = tstore.get_links(entrypath=entry.relpath)[0]
-                    raw_data = toxlink.entry.file_get_content()
-                    data = json.loads(raw_data.decode("utf-8"))
-                    tlink = target_stage.store_toxresult(ref_link, data)
+                    with toxlink.entry.file_open_read() as f:
+                        tlink = target_stage.store_toxresult(ref_link, f)
                     tlink.add_logs(
                         x for x in toxlink.get_logs()
                         if x.get('what') != 'overwrite')
@@ -997,8 +977,8 @@ class PyPIView:
                         dst=target_stage.name)
                     yield (200, "store_toxresult", tlink.entrypath)
         for link in links["doczip"]:
-            doczip = link.entry.file_get_content()
-            new_link = target_stage.store_doczip(name, version, doczip)
+            with link.entry.file_open_read() as doczip:
+                new_link = target_stage.store_doczip(name, version, doczip)
             new_link.add_logs(link.get_logs())
             new_link.add_log(
                 'push',
@@ -1053,13 +1033,12 @@ class PyPIView:
 
                 abort_if_invalid_filename(request, name, content.filename)
                 self._update_versiondata_form(stage, request.POST)
-                file_content = content.file.read()
                 try:
                     link = stage.store_releasefile(
                         project, version,
-                        content.filename, file_content)
+                        content.filename, content.file)
                 except stage.NonVolatile as e:
-                    if e.link.matches_checksum(file_content):
+                    if e.link.matches_checksum(content.file):
                         abort_submit(
                             request, 200,
                             "Upload of identical file to non volatile index.",
@@ -1079,15 +1058,14 @@ class PyPIView:
             else:
                 if "version" in request.POST:
                     self._update_versiondata_form(stage, request.POST)
-                doczip = content.file.read()
                 try:
-                    link = stage.store_doczip(project, version, doczip)
+                    link = stage.store_doczip(project, version, content.file)
                 except stage.MissesVersion as e:
                     abort_submit(
                         request, 400,
                         "%s" % e)
                 except stage.NonVolatile as e:
-                    if e.link.matches_checksum(doczip):
+                    if e.link.matches_checksum(content.file):
                         abort_submit(
                             request, 200,
                             "Upload of identical file to non volatile index.",
@@ -1108,8 +1086,9 @@ class PyPIView:
             if skip_missing and key not in form:
                 continue
             elif key in stage.metadata_list_fields:
-                val = [ensure_unicode(item)
-                        for item in form.getall(key) if item]
+                val = [
+                    ensure_unicode(item)
+                    for item in form.getall(key) if item]
             else:
                 val = form.get(key, "")
                 if val == "UNKNOWN":
@@ -1143,8 +1122,9 @@ class PyPIView:
             stage.set_versiondata(metadata)
         except ValueError as e:
             abort_submit(self.request, 400, "invalid metadata: %s" % (e,))
-        self.log.info("%s: got submit release info %r",
-                 stage.name, metadata["name"])
+        self.log.info(
+            "%s: got submit release info %r",
+            stage.name, metadata["name"])
 
     #
     #  per-project and version data
@@ -1190,8 +1170,8 @@ class PyPIView:
         project = self.context.project
         force = 'force' in self.request.params
         if not stage.ixconfig["volatile"] and not force:
-            apireturn(403, "project %r is on non-volatile index %s" %(
-                      project, stage.name))
+            apireturn(403, "project %r is on non-volatile index %s" % (
+                project, stage.name))
         try:
             stage.del_project(project)
         except KeyError:
@@ -1222,8 +1202,9 @@ class PyPIView:
                 links.append(linkdict)
         shadowing = view_verdata.pop("+shadowing", None)
         if shadowing:
-            view_verdata["+shadowing"] = \
-                    [self._make_view_verdata(x) for x in shadowing]
+            view_verdata["+shadowing"] = [
+                self._make_view_verdata(x)
+                for x in shadowing]
         return view_verdata
 
     @view_config(route_name="/{user}/{index}/{project}/{version}",
@@ -1247,34 +1228,37 @@ class PyPIView:
             relpath = relpath.split("#", 1)[0]
         return relpath
 
-    def _pkgserv(self, entry):
+    def _pkgserv(self, entry, url):
         request = self.request
-        if entry is None:
-            abort(request, 404, "no such file")
-        elif not entry.meta:
+        if not entry.meta:
             abort(request, 410, "file existed, deleted in later serial")
 
         if json_preferred(request):
             entry_data = get_mutable_deepcopy(entry.meta)
             apireturn(200, type="releasefilemeta", result=entry_data)
 
-        if entry.last_modified is None or not entry.file_exists():
-            # The file is in a mirror and either deleted or not yet downloaded.
+        # getting the stage from context will cause 404 if stage is deleted
+        stage = self.context.stage
+
+        file_exists = entry.file_exists()
+        if entry.last_modified is None or not file_exists:
             # We check whether we should serve the file directly
             # or redirect to the external URL
-            stage = self.xom.model.getstage(
-                entry.key.params['user'],
-                entry.key.params['index'])
-            if stage is not None and stage.use_external_url:
-                # redirect to external url
-                return HTTPFound(location=entry.url)
+            if stage.use_external_url:
+                # The file is in a mirror and either deleted or not
+                # yet downloaded. Redirect to external url
+                return HTTPFound(location=URL(url).url)
+            if stage.ixconfig['type'] != "mirror" and not file_exists and not self.xom.is_replica():
+                # return error when private file is missing and not in
+                # replica mode, otherwise fall through to fetch file
+                abort(self.request, 404, "no such file")
 
         if not request.has_permission("pkg_read"):
             abort(request, 403, "package read forbidden")
 
         try:
             if should_fetch_remote_file(entry, request.headers):
-                app_iter = iter_fetch_remote_file(self.xom, entry)
+                app_iter = iter_fetch_remote_file(self.xom, entry, url)
                 headers = next(app_iter)
                 return Response(app_iter=app_iter, headers=headers)
         except BadGateway as e:
@@ -1286,8 +1270,9 @@ class PyPIView:
         if self.request.method == "HEAD":
             return Response(headers=headers)
         else:
-            content = entry.file_get_content()
-            return Response(body=content, headers=headers)
+            # FileIter will close the file
+            return Response(
+                app_iter=FileIter(entry.file_open_read()), headers=headers)
 
     @view_config(route_name="/{user}/{index}/+e/{relpath:.*}")
     def mirror_pkgserv(self):
@@ -1298,16 +1283,24 @@ class PyPIView:
         # they can still recover the deleted release from the filesystem
         # manually in case they need it.
         key = self.xom.filestore.get_key_from_relpath(relpath)
-        if not key.exists():
+        if key is None or not key.exists():
             abort(self.request, 404, "no such file")
         entry = self.xom.filestore.get_file_entry_from_key(key)
-        return self._pkgserv(entry)
+        # we need to add auth back to the url, as aiohttp doesn't include it
+        # in the response url, unlike requests did
+        # we do it in mirror_pkgserv now to avoid storing the credentials
+        # in the database and avoid changes in the db when mirror_url changes
+        mirror_url_auth = getattr(self.context.stage, "mirror_url_auth", {})
+        url = URL(entry.url).replace(**mirror_url_auth)
+        return self._pkgserv(entry, url)
 
     @view_config(route_name="/{user}/{index}/+f/{relpath:.*}")
     def stage_pkgserv(self):
         relpath = self._relpath_from_request()
         entry = self.xom.filestore.get_file_entry(relpath)
-        return self._pkgserv(entry)
+        if entry is None:
+            abort(self.request, 404, "no such file")
+        return self._pkgserv(entry, entry.url)
 
     @view_config(route_name="/{user}/{index}/+e/{relpath:.*}",
                  permission="del_entry",
@@ -1431,8 +1424,8 @@ class PyPIView:
             apireturn(404, "user %r does not exist" % context.username)
         for name, ixconfig in userconfig.get("indexes", {}).items():
             if not ixconfig["volatile"]:
-                apireturn(403, "user %r has non-volatile index: %s" %(
-                               context.username, name))
+                apireturn(403, "user %r has non-volatile index: %s" % (
+                    context.username, name))
         context.user.delete()
         apireturn(200, "user %r deleted" % context.username)
 
@@ -1473,87 +1466,101 @@ def _headers_from_response(r):
     return headers
 
 
-def iter_cache_remote_file(xom, entry):
+def iter_cache_remote_file(xom, entry, url):
     # we get and cache the file and some http headers from remote
-    r = xom.httpget(entry.url, allow_redirects=True)
-    if r.status_code != 200:
-        msg = "error %s getting %s" % (r.status_code, entry.url)
-        threadlog.error(msg)
-        raise BadGateway(msg, code=r.status_code, url=entry.url)
-    threadlog.info("reading remote: %s, target %s", r.url, entry.relpath)
-    content_size = r.headers.get("content-length")
-    err = None
+    running_hash = entry.hash_algo()
+    hash_spec = entry.hash_spec
+    r = xom.httpget(url, allow_redirects=True)
+    f = entry.tx.conn.io_file_new_open(entry._storepath)
+    with contextlib.closing(f):
+        filesize = 0
+        with contextlib.closing(r):
+            if r.status_code != 200:
+                msg = "error %s getting %r" % (r.status_code, url)
+                threadlog.error(msg)
+                raise BadGateway(msg, code=r.status_code, url=url)
+            threadlog.info("reading remote: %r, target %s", URL(r.url), entry.relpath)
+            content_size = r.headers.get("content-length")
+            err = None
 
-    yield _headers_from_response(r)
+            yield _headers_from_response(r)
 
-    content = BytesIO()
-    while 1:
-        data = r.raw.read(10240)
-        if not data:
-            break
-        content.write(data)
-        yield data
+            while 1:
+                data = r.raw.read(10240)
+                if not data:
+                    break
+                filesize += len(data)
+                running_hash.update(data)
+                f.write(data)
+                yield data
 
-    content = content.getvalue()
+        if content_size and int(content_size) != filesize:
+            err = ValueError(
+                "%s: got %s bytes of %r from remote, expected %s" % (
+                    entry.relpath, filesize, r.url, content_size))
+        if not err:
+            err = get_checksum_error(running_hash, entry.relpath, hash_spec)
 
-    filesize = len(content)
-    if content_size and int(content_size) != filesize:
-        err = ValueError(
-            "%s: got %s bytes of %r from remote, expected %s" % (
-                entry.relpath, filesize, r.url, content_size))
-    if not err:
-        err = entry.check_checksum(content)
+        if err is not None:
+            threadlog.error(str(err))
+            raise err
 
-    if err is not None:
-        threadlog.error(str(err))
-        raise err
+        try:
+            # when pushing from a mirror to an index, we are still in a
+            # transaction
+            tx = entry.tx
+        except AttributeError:
+            # when streaming we won't be in a transaction anymore, so we need
+            # to open a new one below
+            tx = None
 
-    try:
-        # when pushing from a mirror to an index, we are still in a
-        # transaction
-        tx = entry.tx
-    except AttributeError:
-        # when streaming we won't be in a transaction anymore, so we need
-        # to open a new one below
-        tx = None
+        def set_content():
+            entry.file_set_content(
+                f, r.headers.get("last-modified", None),
+                hash_spec=hash_spec)
+            if entry.project:
+                stage = xom.model.getstage(
+                    entry.key.params['user'],
+                    entry.key.params['index'])
+                # for mirror indexes this makes sure the project is in the database
+                # as soon as a file was fetched
+                stage.add_project_name(entry.project)
 
-    def set_content():
-        entry.file_set_content(content, r.headers.get("last-modified", None))
-        if entry.project:
-            stage = xom.model.getstage(
-                entry.key.params['user'],
-                entry.key.params['index'])
-            # for mirror indexes this makes sure the project is in the database
-            # as soon as a file was fetched
-            stage.add_project_name(entry.project)
-
-    if not entry.has_existing_metadata():
-        if tx is not None:
-            set_content()
-        else:
-            with xom.keyfs.transaction(write=True):
+        if not entry.has_existing_metadata():
+            if tx is not None:
                 set_content()
-    else:
-        # the file was downloaded before but locally removed, so put
-        # it back in place without creating a new serial
-        # we need a direct write connection to use the io_file_* methods
-        if tx is not None:
-            tx.conn.io_file_set(entry._storepath, content)
-            threadlog.debug(
-                "put missing file back into place: %s", entry._storepath)
+            else:
+                with xom.keyfs.transaction(write=True):
+                    set_content()
+                    # on Windows we need to close the file
+                    # before the transaction closes
+                    f.close()
         else:
-            with xom.keyfs._storage.get_connection(write=True) as conn:
-                conn.io_file_set(entry._storepath, content)
+            # the file was downloaded before but locally removed, so put
+            # it back in place without creating a new serial
+            # we need a direct write connection to use the io_file_* methods
+            if tx is not None:
+                tx.conn.io_file_set(entry._storepath, f)
                 threadlog.debug(
                     "put missing file back into place: %s", entry._storepath)
-                conn.commit_files_without_increasing_serial()
+            else:
+                with xom.keyfs._storage.get_connection(write=True) as conn:
+                    conn.io_file_set(entry._storepath, f)
+                    # on Windows we need to close the file
+                    # before the transaction closes
+                    f.close()
+                    threadlog.debug(
+                        "put missing file back into place: %s", entry._storepath)
+                    conn.commit_files_without_increasing_serial()
 
 
-def iter_remote_file_replica(xom, entry):
+def iter_remote_file_replica(xom, entry, url):
+    running_hash = entry.hash_algo()
+    hash_spec = entry.hash_spec
     replication_errors = xom.replica_thread.shared_data.errors
     # construct master URL with param
-    url = xom.config.master_url.joinpath(entry.relpath).url
-    if not entry.url:
+    master_url = xom.config.master_url.joinpath(entry.relpath).url
+    if not url:
         threadlog.warn("missing private file: %s" % entry.relpath)
     else:
         threadlog.info("replica doesn't have file: %s", entry.relpath)
@@ -1561,73 +1568,77 @@ def iter_remote_file_replica(xom, entry):
     rt = xom.replica_thread
     token = rt.auth_serializer.dumps(uuid)
     r = xom.httpget(
-        url, allow_redirects=True,
+        master_url, allow_redirects=True,
         extra_headers={
             rt.H_REPLICA_FILEREPL: str("YES"),
             rt.H_REPLICA_UUID: uuid,
             str('Authorization'): 'Bearer %s' % token})
     if r.status_code != 200:
-        msg = "%s: received %s from master" % (url, r.status_code)
-        if not entry.url:
+        r.close()
+        msg = "%s: received %s from master" % (master_url, r.status_code)
+        if not url:
             threadlog.error(msg)
             raise BadGateway(msg)
         # try to get from original location
-        r = xom.httpget(entry.url, allow_redirects=True)
-        if r.status_code != 200:
-            msg = "%s\n%s: received %s" % (msg, entry.url, r.status_code)
-            threadlog.error(msg)
-            raise BadGateway(msg)
+        r = xom.httpget(url, allow_redirects=True)
+        with contextlib.closing(r):
+            if r.status_code != 200:
+                msg = "%s\n%s: received %s" % (msg, url, r.status_code)
+                threadlog.error(msg)
+                raise BadGateway(msg)
 
-    yield _headers_from_response(r)
+    f = entry.tx.conn.io_file_new_open(entry._storepath)
+    with contextlib.closing(f):
+        with contextlib.closing(r):
+            yield _headers_from_response(r)
 
-    content = BytesIO()
-    while 1:
-        data = r.raw.read(10240)
-        if not data:
-            break
-        content.write(data)
-        yield data
+            while 1:
+                data = r.raw.read(10240)
+                if not data:
+                    break
+                running_hash.update(data)
+                f.write(data)
+                yield data
 
-    content = content.getvalue()
+        err = get_checksum_error(running_hash, entry.relpath, hash_spec)
+        if err:
+            # the file we got is different, so we fail
+            raise BadGateway(str(err))
 
-    err = entry.check_checksum(content)
-    if err:
-        # the file we got is different, so we fail
-        raise BadGateway(str(err))
-
-    try:
-        # there is no code path that still has a transaction at this point,
-        # but we handle that case just to be safe
-        tx = entry.tx
-    except AttributeError:
-        # when streaming we won't be in a transaction anymore, so we need
-        # to open a new one below
-        tx = None
-    if tx is not None and tx.write:
-        entry.tx.conn.io_file_set(entry._storepath, content)
-    else:
-        # we need a direct write connection to use the io_file_* methods
-        with xom.keyfs._storage.get_connection(write=True) as conn:
-            conn.io_file_set(entry._storepath, content)
-            threadlog.debug(
-                "put missing file back into place: %s", entry._storepath)
-            conn.commit_files_without_increasing_serial()
-    # in case there were errors before, we can now remove them
-    replication_errors.remove(entry)
+        try:
+            # there is no code path that still has a transaction at this point,
+            # but we handle that case just to be safe
+            tx = entry.tx
+        except AttributeError:
+            # when streaming we won't be in a transaction anymore, so we need
+            # to open a new one below
+            tx = None
+        if tx is not None and tx.write:
+            entry.tx.conn.io_file_set(entry._storepath, f)
+        else:
+            # we need a direct write connection to use the io_file_* methods
+            with xom.keyfs._storage.get_connection(write=True) as conn:
+                conn.io_file_set(entry._storepath, f)
+                # on Windows we need to close the file
+                # before the transaction closes
+                f.close()
+                threadlog.debug(
+                    "put missing file back into place: %s", entry._storepath)
+                conn.commit_files_without_increasing_serial()
+        # in case there were errors before, we can now remove them
+        replication_errors.remove(entry)
 
 
-def iter_fetch_remote_file(xom, entry):
+def iter_fetch_remote_file(xom, entry, url):
     filestore = xom.filestore
     keyfs = xom.keyfs
     if not xom.is_replica():
         if not keyfs.tx.write:
             keyfs.restart_as_write_transaction()
         entry = filestore.get_file_entry(entry.relpath, readonly=False)
-        for part in iter_cache_remote_file(xom, entry):
-            yield part
+        yield from iter_cache_remote_file(xom, entry, url)
     else:
-        for part in iter_remote_file_replica(xom, entry):
-            yield part
+        yield from iter_remote_file_replica(xom, entry, url)
 
 
 def url_for_entrypath(request, entrypath):
@@ -1650,11 +1661,12 @@ def getjson(request):
 
 def abort_if_invalid_filename(request, name, filename):
     if not is_valid_archive_name(filename):
-        abort_submit(request, 400, "%r is not a valid archive name" %(filename))
+        abort_submit(request, 400, f"{filename!r} is not a valid archive name")
     if normalize_name(filename).startswith(normalize_name(name)):
         return
-    abort_submit(request, 400, "filename %r does not match project name %r"
-                      %(filename, name))
+    abort_submit(
+        request, 400, "filename %r does not match project name %r" % (
+            filename, name))
 
 
 def abort_if_invalid_project(request, project):
