@@ -95,7 +95,7 @@ def caplog(caplog):
     """ enrich the pytest-catchlog funcarg. """
     def getrecords(msgrex=None, minlevel="DEBUG"):
         if msgrex is not None:
-            msgrex = re.compile(msgrex)
+            msgrex = re.compile(msgrex, re.DOTALL)
         minlevelno = {"DEBUG": 10, "INFO": 20, "WARNING": 30,
                       "ERROR": 40, "FATAL": 50}.get(minlevel)
         recs = []
@@ -280,6 +280,7 @@ def makexom(request, gentmp, httpget, monkeypatch, storage_info):
             # we always need the async_thread
             xom.thread_pool.start_one(xom.async_thread)
         request.addfinalizer(xom.thread_pool.shutdown)
+        request.addfinalizer(xom._close_sessions)
         return xom
     return makexom
 
@@ -356,6 +357,9 @@ def httpget(pypiurls):
                     if "content" in fakeresponse:
                         xself.raw = py.io.BytesIO(fakeresponse["content"])
 
+                def close(self):
+                    return
+
                 @property
                 def status(xself):
                     return xself.status_code
@@ -393,7 +397,9 @@ def httpget(pypiurls):
             headers = kw.setdefault("headers", {})
             if pypiserial is not None:
                 headers["X-PYPI-LAST-SERIAL"] = str(pypiserial)
-            self.mockresponse(remoteurl + name + "/", text=text, **kw)
+            self.mockresponse(
+                URL(remoteurl).joinpath(name).asdir().url,
+                text=text, **kw)
             return ret
 
         def _getmd5digest(self, s):
@@ -1173,7 +1179,6 @@ def _nginx_host_port(host, port, call_devpi_in_dir, server_directory):
         raise
     p = subprocess.Popen([
         nginx, "-c", nginx_conf.strpath, "-p", nginx_directory.strpath])
-    wait_for_port(host, nginx_port)
     return (p, nginx_port)
 
 
@@ -1186,6 +1191,7 @@ def nginx_host_port(request, call_devpi_in_dir, server_directory):
     (p, nginx_port) = _nginx_host_port(
         host, port, call_devpi_in_dir, server_directory)
     try:
+        wait_for_port(host, nginx_port)
         yield (host, nginx_port)
     finally:
         p.terminate()
@@ -1201,6 +1207,7 @@ def nginx_replica_host_port(request, call_devpi_in_dir, server_directory):
     (p, nginx_port) = _nginx_host_port(
         host, port, call_devpi_in_dir, server_directory)
     try:
+        wait_for_port(host, nginx_port)
         yield (host, nginx_port)
     finally:
         p.terminate()
