@@ -7,10 +7,12 @@ try:
 except ImportError:
     from urllib.parse import quote as url_quote
 
+
 class API:
     def __init__(self, d):
         self.__dict__ = d
 
+
 class MappMixin:
     _usercount = 0
 
@@ -408,9 +410,9 @@ class TestMirrorIndexThings:
         r = mapp.downloadrelease((404, 502), result[0])
         msg = r['message']
         assert (
-            'error 404 getting' in msg or
-            'received 404 from master' in msg or
-            'received 502 from master' in msg)
+            'error 404 getting' in msg
+            or 'received 404 from master' in msg
+            or 'received 502 from master' in msg)
 
     def test_download_release(self, mapp, simpypi):
         mapp.create_and_login_user('mirror6')
