@@ -1,11 +1,13 @@
 import py
 import pytest
 
+
 def test_gentmp_suffix_raises(gentmp):
     t = gentmp("source")
     assert not t.listdir()
     pytest.raises(py.error.EEXIST, lambda: gentmp("source"))
 
+
 def test_gentmp_suffix(gentmp):
     source = gentmp("source")
     dest = gentmp("dest")
@@ -13,20 +15,23 @@ def test_gentmp_suffix(gentmp):
     assert source.basename.startswith("source")
     assert dest.basename.startswith("dest")
 
+
 def test_gentmp_unique(gentmp):
     s1 = gentmp()
     s2 = gentmp()
     assert s1 != s2
     assert s1.dirpath() == s2.dirpath()
 
+
 def test_makexom(makexom, maketestapp):
     xom1 = makexom()
     xom2 = makexom()
     assert xom1.config.serverdir != xom2.config.serverdir
     testapp1 = maketestapp(xom1)
-    r = testapp1.get("/")
+    testapp1.get("/")
     testapp2 = maketestapp(xom2)
-    r = testapp2.get("/")
+    testapp2.get("/")
+
 
 def test_makemapp(makemapp):
     mapp1 = makemapp()
