@@ -124,6 +124,7 @@ class TestFileStore:
 
     @pytest.mark.storage_with_filesystem
     def test_file_exists_new_hash(self, filestore, gen, xom):
+        from devpi_server.filestore import get_checksum_error
         content1 = b'somedata'
         md5_1 = hashlib.md5(content1).hexdigest()
         link1 = gen.pypi_package_link("pytest-1.2.zip", md5=md5_1)
@@ -140,8 +141,8 @@ class TestFileStore:
         assert entry2.file_exists()
         content2 = entry2.file_get_content()
         assert content1 != content2
-        assert isinstance(entry2.check_checksum(content2), ValueError)
-        assert entry2.check_checksum(content1) is None
+        assert isinstance(get_checksum_error(content2, entry2.relpath, entry2.hash_spec), ValueError)
+        assert get_checksum_error(content1, entry2.relpath, entry2.hash_spec) is None
         filestore.keyfs.commit_transaction_in_thread()
         assert py.path.local(filepath).exists()
 
@@ -182,7 +183,7 @@ class TestFileStore:
         headers = ResponseHeaders({})
         httpget.url2response[link.url] = dict(status_code=200,
                 headers=headers, raw = BytesIO(b"123"))
-        for part in iter_cache_remote_file(xom, entry):
+        for part in iter_cache_remote_file(xom, entry, entry.url):
             pass
         rheaders = entry.gethttpheaders()
         assert rheaders["content-length"] == "3"
@@ -196,7 +197,7 @@ class TestFileStore:
         headers = ResponseHeaders({"Content-Type": ""})
         httpget.url2response[link.url] = dict(status_code=200,
                 headers=headers, raw = BytesIO(b"123"))
-        for part in iter_cache_remote_file(xom, entry):
+        for part in iter_cache_remote_file(xom, entry, entry.url):
             pass
         rheaders = entry.gethttpheaders()
         assert rheaders["content-length"] == "3"
@@ -214,7 +215,7 @@ class TestFileStore:
         httpget.url2response[link.url] = dict(status_code=200,
                 headers=headers, raw = BytesIO(b"1"))
         with pytest.raises(ValueError):
-            for part in iter_cache_remote_file(xom, entry):
+            for part in iter_cache_remote_file(xom, entry, entry.url):
                 pass
 
     def test_iterfile_remote_nosize(self, filestore, httpget, gen, xom):
@@ -227,7 +228,7 @@ class TestFileStore:
         assert entry.file_size() is None
         httpget.url2response[link.url] = dict(status_code=200,
                 headers=headers, raw=BytesIO(b"1"))
-        for part in iter_cache_remote_file(xom, entry):
+        for part in iter_cache_remote_file(xom, entry, entry.url):
             pass
         assert entry.file_get_content() == b"1"
         entry2 = filestore.get_file_entry(entry.relpath)
@@ -247,7 +248,7 @@ class TestFileStore:
         httpget.url2response[link.url_nofrag] = dict(status_code=200,
                 headers=headers, raw=BytesIO(b"123"))
         with pytest.raises(ValueError, match=link.md5):
-            for part in iter_cache_remote_file(xom, entry):
+            for part in iter_cache_remote_file(xom, entry, entry.url):
                 pass
         assert not entry.file_exists()
 
@@ -265,7 +266,7 @@ def test_cache_remote_file(filestore, httpget, gen, xom):
             status_code=200,
             headers=headers,
             raw=BytesIO(b"123"))
-        for part in iter_cache_remote_file(xom, entry):
+        for part in iter_cache_remote_file(xom, entry, entry.url):
             pass
         rheaders = entry.gethttpheaders()
         assert rheaders["content-length"] == "3"
