@@ -13,7 +13,7 @@ import os
 
 @pytest.fixture
 def ground_wsgi_run(monkeypatch):
-    monkeypatch.setattr(mythread.ThreadPool, "live", lambda *args: 0 / 0)
+    monkeypatch.setattr(mythread.ThreadPool, "run", lambda *args: 0 / 0)
 
 
 wsgi_run_throws = pytest.mark.usefixtures("ground_wsgi_run")
