@@ -346,6 +346,8 @@ class TestProxyViewToMaster:
         request = blank_request(method="PUT")
         request.registry = dict(xom=xom)
         response = proxy_view_to_master(None, request)
+        # exhaust app_iter to close the proxied response
+        list(response.app_iter)
         assert response.headers.get("X-DEVPI-SERIAL") == "10"
         assert l == [10]
 
@@ -359,6 +361,8 @@ class TestProxyViewToMaster:
         request = blank_request(method="PUT")
         request.registry = dict(xom=xom)
         response = proxy_view_to_master(None, request)
+        # exhaust app_iter to close the proxied response
+        list(response.app_iter)
         assert response.status == "200 GOOD"
 
     def test_write_proxies_redirect(self, makexom, blank_request, reqmock, monkeypatch):
@@ -375,6 +379,8 @@ class TestProxyViewToMaster:
         request = blank_request(method="PUT", headers=dict(host='my.domain'))
         request.registry = dict(xom=xom)
         response = proxy_view_to_master(None, request)
+        # exhaust app_iter to close the proxied response
+        list(response.app_iter)
         assert response.headers.get("X-DEVPI-SERIAL") == "10"
         assert response.headers.get("location") == "http://my.domain/hello"
         assert l == [10]
@@ -393,6 +399,8 @@ class TestProxyViewToMaster:
         request = blank_request(method="PUT")
         request.registry = dict(xom=xom)
         response = proxy_view_to_master(None, request)
+        # exhaust app_iter to close the proxied response
+        list(response.app_iter)
         assert 'connection' not in response.headers
         assert 'foo' not in response.headers
         assert 'keep-alive' not in response.headers
@@ -413,10 +421,10 @@ def replay(xom, replica_xom, events=True):
             change_entry = conn.get_changes(serial)
         threadlog.info("test: importing to replica %s", serial)
         replica_xom.keyfs.import_changes(serial, change_entry)
+    replica_xom.replica_thread.wait()
 
     # replay notifications
     if events:
-        replica_xom.replica_thread.wait()
         noti_thread = replica_xom.keyfs.notifier
         event_serial = noti_thread.read_event_serial()
         thread_push_log("NOTI")
@@ -436,7 +444,6 @@ def make_replica_xom(makexom, secretfile):
             "--secretfile", secretfile.strpath] + list(options))
         # shorten error delay for tests
         replica_xom.replica_thread.shared_data.ERROR_QUEUE_MAX_DELAY = 0.1
-        replica_xom.thread_pool.start_one(replica_xom.replica_thread)
         replica_xom.thread_pool.start_one(
             replica_xom.replica_thread.file_replication_threads[0])
         return replica_xom
@@ -462,6 +469,7 @@ class TestUseExistingFiles:
             '--replica-file-search-path', existing_base.strpath])
         # now sync the replica, if the file is not found there will be an error
         # because httpget is mocked
+        caplog.clear()
         replay(xom, replica_xom, events=False)
         replica_xom.replica_thread.wait()
         assert len(caplog.getrecords('checking existing file')) == 1
@@ -489,6 +497,7 @@ class TestUseExistingFiles:
             '--hard-links'])
         # now sync the replica, if the file is not found there will be an error
         # because httpget is mocked
+        caplog.clear()
         replay(xom, replica_xom)
         assert len(caplog.getrecords('checking existing file')) == 1
         # check the number of links of the file
@@ -515,6 +524,7 @@ class TestUseExistingFiles:
         frt_reqmock.mockresponse("http://localhost" + path, 200, data=content1)
         # now sync the replica, if the file is not found there will be an error
         # because httpget is mocked
+        caplog.clear()
         replay(xom, replica_xom, events=False)
         replica_xom.replica_thread.wait()
         assert len(caplog.getrecords('checking existing file')) == 1
@@ -549,6 +559,7 @@ class TestUseExistingFiles:
         frt_reqmock.mockresponse("http://localhost" + path, 200, data=content1)
         # now sync the replica, if the file is not found there will be an error
         # because httpget is mocked
+        caplog.clear()
         replay(xom, replica_xom)
         assert len(caplog.getrecords('checking existing file')) == 1
         assert len(caplog.getrecords('sha256 mismatch, got ec7d057f450dc963f15978af98b9cdda64aca6751c677e45d4a358fe103dc05b')) == 1
@@ -741,7 +752,7 @@ class TestFileReplication:
             url = replica_xom.config.master_url.joinpath(entry.relpath).url
             pypistage.xom.httpget.mockresponse(url, status_code=500)
             with pytest.raises(BadGateway) as e:
-                for part in iter_remote_file_replica(replica_xom, entry):
+                for part in iter_remote_file_replica(replica_xom, entry, entry.url):
                     pass
             e.match('received 500 from master')
             e.match('pypi.org/package/some/pytest-1.8.zip: received 404')
@@ -769,7 +780,7 @@ class TestFileReplication:
             pypistage.xom.httpget.mockresponse(url, status_code=500)
             pypistage.xom.httpget.mockresponse(
                 entry.url, headers=headers, content=b'123')
-            result = iter_remote_file_replica(replica_xom, entry)
+            result = iter_remote_file_replica(replica_xom, entry, entry.url)
             headers = next(result)
             # there should be one get
             (call_log_entry,) = [
@@ -826,7 +837,6 @@ class TestFileReplication:
 
 
 def test_simplelinks_update_updates_projectname(pypistage, replica_xom, xom):
-    replica_xom.thread_pool.start_one(replica_xom.replica_thread)
     replica_xom.thread_pool.start_one(
         replica_xom.replica_thread.file_replication_threads[0])
     pypistage.mock_simple_projects([])
@@ -847,7 +857,6 @@ def test_simplelinks_update_updates_projectname(pypistage, replica_xom, xom):
 
 def test_get_simplelinks_perstage(httpget, monkeypatch, pypistage, replica_pypistage,
                                   pypiurls, replica_xom, xom):
-    replica_xom.thread_pool.start_one(replica_xom.replica_thread)
     replica_xom.thread_pool.start_one(
         replica_xom.replica_thread.file_replication_threads[0])
 
@@ -890,16 +899,26 @@ def test_get_simplelinks_perstage(httpget, monkeypatch, pypistage, replica_pypis
     assert xom.keyfs.get_current_serial() > serial
 
     # we patch wait_tx_serial so we can check it
-    orig_wait_tx_serial = replica_xom.keyfs.wait_tx_serial
     called = []
 
     def wait_tx_serial(serial, timeout=None):
-        result = orig_wait_tx_serial(serial, timeout=timeout)
-        called.append((serial, result))
-        return result
+        threadlog.info("Patched wait_tx_serial %s", serial)
+        called.append(serial)
+        assert xom.keyfs.get_current_serial() == serial
+        assert replica_xom.keyfs.get_current_serial() < serial
+        # save _import_subscriber
+        import_subscriber = replica_xom.keyfs._import_subscriber
+        # set _import_subscriber to empty dict, so no transaction is opened
+        # during import_changes
+        replica_xom.keyfs._import_subscriber = {}
+        replay(xom, replica_xom, events=False)
+        # restore _import_subscriber
+        replica_xom.keyfs._import_subscriber = import_subscriber
+        replica_xom.keyfs.restart_read_transaction()
+        assert replica_xom.keyfs.get_current_serial() == serial
+        return True
 
     monkeypatch.setattr(replica_xom.keyfs, 'wait_tx_serial', wait_tx_serial)
-    replay(xom, replica_xom, events=False)
     pypiurls.simple = 'http://localhost:3111/root/pypi/+simple/'
     httpget.mock_simple(
         'pytest',
@@ -911,14 +930,13 @@ def test_get_simplelinks_perstage(httpget, monkeypatch, pypistage, replica_pypis
         r_pypistage = replica_xom.model.getstage("root/pypi")
         r_pypistage.cache_retrieve_times.expire("pytest")
         ret = replica_pypistage.get_releaselinks("pytest")
-    assert called == [(2, True)]
+    assert called == [2]
     replay(xom, replica_xom)
     assert len(ret) == 1
     assert ret[0].relpath == 'root/pypi/+e/https_pypi.org_pytest/pytest-1.1.zip'
 
 
 def test_replicate_deleted_user(mapp, replica_xom):
-    replica_xom.thread_pool.start_one(replica_xom.replica_thread)
     replica_xom.thread_pool.start_one(
         replica_xom.replica_thread.file_replication_threads[0])
     mapp.create_and_use("hello/dev")
@@ -1004,37 +1022,50 @@ class TestFileReplicationSharedData:
         # allow on_import to run right away, so we don't need to rely
         # on the initial import thread for tests
         replica_xom.replica_thread.replica_in_sync_at = 0
-        return FileReplicationSharedData(replica_xom)
+        shared_data = FileReplicationSharedData(replica_xom)
+        # No waiting on empty queues
+        shared_data.QUEUE_TIMEOUT = 0.1
+        return shared_data
 
     def test_mirror_priority(self, shared_data):
+        from itertools import permutations
         result = []
 
         mirror_file = 'root/pypi/+f/3f8/3058ac9076112/pytest-2.0.0.zip'
         stage_file = 'root/dev/+f/274/e88b0b3d028fe/pytest-2.1.0.zip'
+        deleted_file = 'root/dev1/+f/274/e88b0b3d028fe/pytest-2.2.0.zip'
         # set the index_types cache to prevent db access
-        shared_data.index_types.put('root/pypi', 'mirror')
-        shared_data.index_types.put('root/dev', 'stage')
+        shared_data.set_index_type_for('root/pypi', 'mirror')
+        shared_data.set_index_type_for('root/dev', 'stage')
+        shared_data.set_index_type_for('root/dev1', None)
 
-        def handler(is_from_mirror, serial, key, keyname, value, back_serial):
+        def handler(index_type, serial, key, keyname, value, back_serial):
             result.append(key)
         # Regardless of the serial or add order, the stage should come first
-        cases = [
-            ((mirror_file, 0), (stage_file, 0)),
-            ((mirror_file, 1), (stage_file, 0)),
-            ((mirror_file, 0), (stage_file, 1)),
-            ((stage_file, 0), (mirror_file, 0)),
-            ((stage_file, 1), (mirror_file, 0)),
-            ((stage_file, 0), (mirror_file, 1))]
-        for (relpath1, serial1), (relpath2, serial2) in cases:
+        cases = []
+        cases.extend(permutations(((mirror_file, 0), (stage_file, 0), (deleted_file, 0))))
+        cases.extend(permutations(((mirror_file, 1), (stage_file, 0), (deleted_file, 0))))
+        cases.extend(permutations(((mirror_file, 0), (stage_file, 1), (deleted_file, 0))))
+        cases.extend(permutations(((mirror_file, 0), (stage_file, 0), (deleted_file, 1))))
+        cases.extend(permutations(((mirror_file, 0), (stage_file, 1), (deleted_file, 2))))
+        cases.extend(permutations(((mirror_file, 0), (stage_file, 2), (deleted_file, 1))))
+        cases.extend(permutations(((mirror_file, 1), (stage_file, 0), (deleted_file, 2))))
+        cases.extend(permutations(((mirror_file, 1), (stage_file, 2), (deleted_file, 0))))
+        cases.extend(permutations(((mirror_file, 2), (stage_file, 0), (deleted_file, 1))))
+        cases.extend(permutations(((mirror_file, 2), (stage_file, 1), (deleted_file, 0))))
+        for (relpath1, serial1), (relpath2, serial2), (relpath3, serial3) in cases:
             key1 = shared_data.xom.keyfs.get_key_instance('STAGEFILE', relpath1)
             key2 = shared_data.xom.keyfs.get_key_instance('STAGEFILE', relpath2)
-            shared_data.on_import(None, serial1, key1, None, -1)
-            shared_data.on_import(None, serial2, key2, None, -1)
-            assert shared_data.queue.qsize() == 2
+            key3 = shared_data.xom.keyfs.get_key_instance('STAGEFILE', relpath3)
+            shared_data.on_import_file(None, serial1, key1, None, -1)
+            shared_data.on_import_file(None, serial2, key2, None, -1)
+            shared_data.on_import_file(None, serial3, key3, None, -1)
+            assert shared_data.queue.qsize() == 3
+            shared_data.process_next(handler)
             shared_data.process_next(handler)
             shared_data.process_next(handler)
             assert shared_data.queue.qsize() == 0
-            assert result == [stage_file, mirror_file]
+            assert result == [stage_file, mirror_file, deleted_file]
             result.clear()
 
     @pytest.mark.parametrize("index_type", ["mirror", "stage"])
@@ -1042,16 +1073,16 @@ class TestFileReplicationSharedData:
         relpath = 'root/dev/+f/274/e88b0b3d028fe/pytest-2.1.0.zip'
         key = shared_data.xom.keyfs.get_key_instance('STAGEFILE', relpath)
         # set the index_types cache to prevent db access
-        shared_data.index_types.put('root/dev', 'stage')
+        shared_data.set_index_type_for('root/dev', 'stage')
         result = []
 
-        def handler(is_from_mirror, serial, key, keyname, value, back_serial):
+        def handler(index_type, serial, key, keyname, value, back_serial):
             result.append(serial)
 
         # Later serials come first
-        shared_data.on_import(None, 1, key, None, -1)
-        shared_data.on_import(None, 100, key, None, -1)
-        shared_data.on_import(None, 10, key, None, -1)
+        shared_data.on_import_file(None, 1, key, None, -1)
+        shared_data.on_import_file(None, 100, key, None, -1)
+        shared_data.on_import_file(None, 10, key, None, -1)
         assert shared_data.queue.qsize() == 3
         shared_data.process_next(handler)
         shared_data.process_next(handler)
@@ -1063,7 +1094,7 @@ class TestFileReplicationSharedData:
         relpath = 'root/dev/+f/274/e88b0b3d028fe/pytest-2.1.0.zip'
         key = shared_data.xom.keyfs.get_key_instance('STAGEFILE', relpath)
         # set the index_types cache to prevent db access
-        shared_data.index_types.put('root/dev', 'stage')
+        shared_data.set_index_type_for('root/dev', 'stage')
 
         next_ts_result = []
         handler_result = []
@@ -1074,13 +1105,11 @@ class TestFileReplicationSharedData:
             return orig_next_ts(delay)
         shared_data.next_ts = next_ts
 
-        def handler(is_from_mirror, serial, key, keyname, value, back_serial):
+        def handler(index_type, serial, key, keyname, value, back_serial):
             handler_result.append(key)
             raise ValueError
 
-        # No waiting on empty queues
-        shared_data.QUEUE_TIMEOUT = 0
-        shared_data.on_import(None, 0, key, None, -1)
+        shared_data.on_import_file(None, 0, key, None, -1)
         assert shared_data.queue.qsize() == 1
         assert shared_data.error_queue.qsize() == 0
         assert next_ts_result == []
@@ -1121,3 +1150,88 @@ class TestFileReplicationSharedData:
         # or ERROR_QUEUE_MAX_DELAY is changed
         assert len(next_ts_result) == 17
         assert len(handler_result) == 17
+
+    def test_deleted_index(self, monkeypatch, shared_data):
+        from devpi_server.replica import IndexType
+        relpath = "root/dev/+f/274/e88b0b3d028fe/pytest-2.1.0.zip"
+        key = shared_data.xom.keyfs.get_key_instance("STAGEFILE", relpath)
+
+        result = []
+
+        def handler(index_type, serial, key, keyname, value, back_serial):
+            result.append((index_type, serial, key, keyname, value, back_serial))
+
+        # simulate deleted index
+        l = []
+        monkeypatch.setattr(
+            shared_data.xom.model, "getstage", lambda u, i: l.append((u, i)))
+        shared_data.on_import_file(None, 1, key, None, -1)
+        # it should still be queued to check master and get a 410 for sure
+        assert shared_data.queue.qsize() == 1
+        shared_data.process_next(handler)
+        ((index_type, serial, key, keyname, value, back_serial),) = result
+        assert index_type == IndexType(None)
+        assert key == relpath
+
+    def test_recreated_index(self, monkeypatch, shared_data):
+        from devpi_server.replica import IndexType
+        relpath = "root/dev/+f/274/e88b0b3d028fe/pytest-2.1.0.zip"
+        key = shared_data.xom.keyfs.get_key_instance("STAGEFILE", relpath)
+        userkey = shared_data.xom.keyfs.get_key_instance("USER", "root/.config")
+
+        result = []
+
+        def handler(index_type, serial, key, keyname, value, back_serial):
+            result.append(index_type)
+
+        changes = {
+            0: {userkey: ({"indexes": {"dev": {"type": "stage"}}}, -1)},
+            1: {key: (None, -1)},
+            2: {userkey: ({}, 0)},
+            3: {key: (None, -1)},
+            4: {userkey: ({"indexes": {"dev": {"type": "mirror"}}}, 2)},
+            5: {key: (None, -1)},
+            6: {userkey: (None, 4)},
+            7: {key: (None, -1)}}
+
+        class Tx:
+            def get_value_at(self, key, serial):
+                return changes[serial][key][0]
+
+        monkeypatch.setattr("devpi_server.keyfs.KeyFS.tx", Tx())
+
+        # simulate index creation
+        shared_data.on_import(0, changes[0])
+        # import file
+        shared_data.on_import(1, changes[1])
+        assert shared_data.queue.qsize() == 1
+        shared_data.process_next(handler)
+        (index_type,) = result
+        assert index_type == IndexType("stage")
+        result.clear()
+        # simulate index deletion
+        shared_data.on_import(2, changes[2])
+        # import file
+        shared_data.on_import(3, changes[3])
+        assert shared_data.queue.qsize() == 1
+        shared_data.process_next(handler)
+        (index_type,) = result
+        assert index_type == IndexType(None)
+        result.clear()
+        # simulate index recreation
+        shared_data.on_import(4, changes[4])
+        # import file
+        shared_data.on_import(5, changes[5])
+        assert shared_data.queue.qsize() == 1
+        shared_data.process_next(handler)
+        (index_type,) = result
+        assert index_type == IndexType("mirror")
+        result.clear()
+        # simulate user deletion
+        shared_data.on_import(6, changes[6])
+        # import file
+        shared_data.on_import(7, changes[7])
+        assert shared_data.queue.qsize() == 1
+        shared_data.process_next(handler)
+        (index_type,) = result
+        assert index_type == IndexType(None)
