@@ -1,4 +1,5 @@
 import base64
+import contextlib
 import json
 import pytest
 import requests
@@ -34,7 +35,8 @@ def server_url_session(host_port, simpypi):
         r = s.put(url + 'root/mirror', json.dumps(indexconfig)).json()
         assert r['type'] == 'indexconfig'
         assert r['result']['mirror_url'] == simpypi.simpleurl
-    return (url, s)
+    yield (url, s)
+    s.close()
 
 
 @pytest.fixture(scope="session")
@@ -53,31 +55,35 @@ def files_directory(server_directory):
 
 
 class TestStreaming(object):
-    @pytest.mark.parametrize("length,pkg_version", [
-        (None, '1.0'), (False, '1.1')])
-    def test_streaming_download(self, content_digest, files_directory, length, pkg_version, server_url_session, simpypi, storage_info):
+    @pytest.mark.parametrize("length,pkg_version,pkg_name", [
+        (None, '1.0', 'pkg1'), (False, '1.1', 'pkg2')])
+    def test_streaming_download(self, content_digest, files_directory, length, pkg_version, pkg_name, server_url_session, simpypi, storage_info):
         from time import sleep
         if "storage_with_filesystem" not in storage_info.get('_test_markers', []):
             pytest.skip("The storage doesn't have marker 'storage_with_filesystem'.")
         (content, digest) = content_digest
         (url, s) = server_url_session
-        pkgzip = "pkg-%s.zip" % pkg_version
-        simpypi.add_release('pkg', pkgver='%s#sha256=%s' % (pkgzip, digest))
-        simpypi.add_file('/pkg/%s' % pkgzip, content, stream=True, length=length)
-        r = s.get(url + 'root/mirror/pkg').json()
+        pkgzip = f"{pkg_name}-{pkg_version}.zip"
+        simpypi.add_release(pkg_name, pkgver='%s#sha256=%s' % (pkgzip, digest))
+        simpypi.add_file(
+            f"/{pkg_name}/{pkgzip}", content, stream=True, length=length)
+        with contextlib.closing(s.get(url + f"root/mirror/{pkg_name}")) as r:
+            r = r.json()
+        assert pkg_version in r['result'], r
         href = r['result'][pkg_version]['+links'][0]['href']
         r = requests.get(href, stream=True)
-        stream = r.iter_content(1024)
-        data = next(stream)
-        assert data == b'deadbeaf' * 128
-        part = next(stream)
-        assert part == b'sandwich' * 128
-        data = data + part
-        if length is not False:
-            assert r.headers['content-length'] == str(len(content))
-        for part in stream:
+        with contextlib.closing(r):
+            stream = r.iter_content(1024)
+            data = next(stream)
+            assert data == b'deadbeaf' * 128
+            part = next(stream)
+            assert part == b'sandwich' * 128
             data = data + part
-        assert data == content
+            if length is not False:
+                assert r.headers['content-length'] == str(len(content))
+            for part in stream:
+                data = data + part
+            assert data == content
         pkg_file = files_directory.join(
             'root', 'mirror', '+f', digest[:3], digest[3:16], pkgzip)
         # this is sometimes delayed a bit, so we check for a while
@@ -87,30 +93,34 @@ class TestStreaming(object):
             sleep(0.1)
         assert pkg_file.exists()
 
-    @pytest.mark.parametrize("size_factor,pkg_version", [
-        (2, '1.2'), (0.5, '1.3')])
-    def test_streaming_differing_content_size(self, content_digest, files_directory, pkg_version, server_url_session, simpypi, size_factor, storage_info):
+    @pytest.mark.parametrize("size_factor,pkg_version,pkg_name", [
+        (2, '1.2', 'pkg3'), (0.5, '1.3', 'pkg4')])
+    def test_streaming_differing_content_size(self, content_digest, files_directory, pkg_version, pkg_name, server_url_session, simpypi, size_factor, storage_info):
         if "storage_with_filesystem" not in storage_info.get('_test_markers', []):
             pytest.skip("The storage doesn't have marker 'storage_with_filesystem'.")
         (content, digest) = content_digest
         (url, s) = server_url_session
-        pkgzip = "pkg-%s.zip" % pkg_version
+        pkgzip = f"{pkg_name}-{pkg_version}.zip"
         length = int(len(content) * size_factor)
-        simpypi.add_release('pkg', pkgver='%s#sha256=%s' % (pkgzip, digest))
-        simpypi.add_file('/pkg/%s' % pkgzip, content, stream=True, length=length)
-        r = s.get(url + 'root/mirror/pkg').json()
+        simpypi.add_release(pkg_name, pkgver='%s#sha256=%s' % (pkgzip, digest))
+        simpypi.add_file(
+            f"/{pkg_name}/{pkgzip}", content, stream=True, length=length)
+        with contextlib.closing(s.get(url + f"root/mirror/{pkg_name}")) as r:
+            r = r.json()
+        assert pkg_version in r['result'], r
         href = r['result'][pkg_version]['+links'][0]['href']
         r = requests.get(href, stream=True)
-        stream = r.iter_content(1024)
-        data = next(stream)
-        assert data == b'deadbeaf' * 128
-        part = next(stream)
-        assert part == b'sandwich' * 128
-        data = data + part
-        assert r.headers['content-length'] == str(length)
-        for part in stream:
+        with contextlib.closing(r):
+            stream = r.iter_content(1024)
+            data = next(stream)
+            assert data == b'deadbeaf' * 128
+            part = next(stream)
+            assert part == b'sandwich' * 128
             data = data + part
-        assert data == content[:length]
+            assert r.headers['content-length'] == str(length)
+            for part in stream:
+                data = data + part
+            assert data == content[:length]
         pkg_file = files_directory.join(
             'root', 'pypi', '+f', digest[:3], digest[3:16], pkgzip)
         assert not pkg_file.exists()
